
"""
======================================================
Visualization package for LomPy (:mod:`lompy.viz`)
======================================================
          
.. currentmodule: LomPy.viz

This subpackage includes vizualization routines built on NumPy, GeoPandas and Matplotlib

Routine 1:
    
    LomPy.viz.img_arrayVD
    LomPy.viz.img_arrayRD
    
    
Routine 2:
    
    LomPy.viz.global_scalingVD
    LomPy.viz.global_scalingRD

Routine 3:
    
    LomPy.viz.plot_histo_VD
    LomPy.viz.plot_histo_RD
    
Routine 4:
    LomPy.viz.plot_gdf_fit
    LomPy.viz.plot_gdf_radius
"""

#Copyright (C) 2023 GNU AGPLv3, LomPy 2023, S.G. Roux and J. Lengyel.
#All rights reserved.
#Contact: stephane.roux@ens-lyon.fr, jankalengyel@gmail.com
#Other Contributors: P. Thiraux

#%%
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import sys
#sys.path.append("../aux/")##!!tag pipchange
#import aux as aux ##!!tag pipchange
import lompy.aux as aux
import geopandas as gpd
#%%


def img_arrayVD(data, X, Y, figsize =(8,8) , cmap = 'Spectral', vmin = np.array([]), vmax = np.array([]), colorbar = False):
    """Return image object from 1D array   
      
    Parameters
    ---------------------
    data : numpy.array
        The values to be visualized (1D)   
    X,Y : numpy.ndarray
        The grid resolution, X and Y must have the same size 
    figsize: tuple with two elements, optional
        The size of the final image
    cmap: matplotlib.colors.Colormap object, optional
        Default is using "Spectral"
    vmin, vmax: float, optional
        Parameters to construct a matplotlib.colors.Normalize()
        Default: np.nanquantile(data,0.015), np.nanquantile(data,0.985)
    colorbar: boolean, optional 
        if True it displays a matplotlib.pyplot.colorbar, default False
        
    Returns
    ______________________
    im : 2D image with scalar data
                            
    Example
    ---------------------
    im = img_arrayVD(data, X, Y)     
    
    Notes
    ---------------------
    matplotlib.pyplot.imshow is used to return a 2D image with scalar 
    data - of a predefined resolution - using a 1D numpy array as input
    """ 
    
    
    if data.shape[0] == X.shape[0]*X.shape[1]:
        if isinstance(vmin, (np.floating, float)) == False:
            vmin = np.nanquantile(data,0.015)
            vmax = np.nanquantile(data,0.985)
        im = plt.imshow(np.reshape(data,(X.shape[0],Y.shape[1])), cmap = cmap, origin = 'lower', vmin = vmin, vmax=vmax)
        plt.xticks([], [])
        plt.yticks([], [])
        if colorbar == True:
            plt.colorbar()
        return im
    else:
        print('data and XY must have the same shape')
    
#%%    
    
    
def img_arrayRD(data, figsize =(8,8) , cmap = 'Spectral', vmin = np.array([]), vmax = np.array([]), colorbar = False):
    """Return image object from 2D scalar data   
      
    Parameters
    ---------------------
    data : numpy.narray
        The values to be visualized (2D scalar data)   
    figsize: tuple with two elements, optional
        The size of the final image
    cmap: matplotlib.colors.Colormap object, optional
        Default is using "Spectral"
    vmin, vmax: float, optional
        Parameters to construct a matplotlib.colors.Normalize()
        Default: np.nanquantile(data,0.015), np.nanquantile(data,0.985)
    colorbar: boolean, optional 
        if True it displays a matplotlib.pyplot.colorbar, default False
        
    Returns
    ______________________
    im : 2D image with scalar data
                            
    Example
    ---------------------
    im = img_array2D(data)     
    
    Notes
    ---------------------
    matplotlib.pyplot.imshow is used to return a 2D image with scalar 
    data - of a predefined resolution - using a 1D numpy array as input
    """ 
    
    
    if len(data.shape) == 2:
        if isinstance(vmin, (np.floating, float)) == False:
            vmin = np.nanquantile(data,0.015)
            vmax = np.nanquantile(data,0.985)
        im = plt.imshow(data, cmap = cmap, origin = 'lower', vmin = vmin, vmax=vmax)
        plt.xticks([], [])
        plt.yticks([], [])
        if colorbar == True:
            plt.colorbar()
        return im
    else:
        print('data must be 2D scalar data')
#%%
def global_scalingVD(data, radius, ml=[], label=[] , color = 'black', alpha = 0.9, linewidth = 1., markersize=3.75):
    """Return a plot of global scaling from vector data results (KNN) 
    
    Parameters
    ---------------------
    data : numpy.narray
        The values to be visualized 
    radius: numpy.array
        The radii used for the multiscale analysis
    cmap: matplotlib.colors.Colormap object, optional
        Default is using "Spectral"
    alpha: float, optional
    color: string, optional
    linewidth: float, optional
        
    Returns
    ______________________
    matplotlib line plot of global scaling
                            
    Example
    ---------------------
    global_scalingVD(data, radius)

    Notes
    _____________________
    matplotlib.pyplot.plot is used to plot line, y versus x  
    """
    if data.shape[1] == len(radius):
        tmp=np.nanmean(data,axis=0)         
        plt.plot(np.log(radius),(tmp.T), ml , label=label, color = color, alpha=alpha, linewidth=linewidth, markersize = markersize)
    else:
        print('The second dimension of the data must be equal to the length of radii')


def global_scalingRD(data, radius, ml=[], label=[] , color = 'black', alpha = 0.9, linewidth = 1., markersize=5.):
    """Return a plot of global scaling from raster data results (conv2D)
    
    Parameters
    ---------------------
    data : numpy.narray
        The scalar data to be visualized 
    radius: numpy.array
        The radii used for the multiscale analysis
    cmap: matplotlib.colors.Colormap object, optional
        Default is using "Spectral"
    alpha: float, optional
    color: color, optional
    linewidth: float, optional
        
    Returns
    ______________________
    matplotlib line plot of global scaling
                            
    Example
    ---------------------
    global_scalingVD(data, radius)

    Notes
    _____________________
    matplotlib.pyplot.plot is used to plot line, y versus x  
    """
    if data.shape[2] == len(radius):
        tmp =  np.nanmean(data.reshape(-1,len(radius)), axis =0)             
        plt.plot(np.log(radius),(tmp.T), ml , label=label, color = color, alpha=alpha, linewidth=linewidth, markersize = markersize)
    else: 
        print('The third dimension of the data must be equal to the length of radii')
#%%

def gplot_df_fit(data, fit, grid, radius, figsize =(20,20) , scheme="quantiles", k=6, cmap = 'Spectral', edgecolor = 'face', legend = False):
    """Visualize LomPy scaling exponents on maps
    """
    gdf = aux.build_gdf(grid, radius, data)
    if data.shape[1] == len(radius):
        gdf['fit'] = fit
        fig, ax =plt.figure(figsize=figsize)
        gdf.plot(column=gdf.fit, scheme=scheme,  k=k, ax = ax, cmap = cmap, edgecolor = edgecolor, legend = legend)
        plt.axis('off')
        plt.tight_layout()
        if legend == True:
            plt.legend()
    else:
        print('The second dimension of the data must be equal to the length of radii')

#%%

def plot_gdf_radius(data, fit, grid, radius, vizradius, figsize =(20,20) , scheme="quantiles", k=6, cmap = 'Spectral', edgecolor = 'face', legend = False):
    """Visualize LomPy scaling functions on maps
    """
    gdf = aux.build_gdf(grid, radius, data)
    if data.shape[1] == len(radius):
        gdf['rr'] = gdf.loc[vizradius]
        fig, ax =plt.figure(figsize=figsize)
        gdf.plot(column=gdf.rr, scheme=scheme,  k=k, ax = ax, cmap = cmap, edgecolor = edgecolor, legend = legend)
        plt.axis('off')
        plt.tight_layout()
        if legend == True:
            plt.legend()
    else:
        print('The second dimension of the data must be equal to the length of radii')
#%%
def plot_histo_VD(data, bins = 50, color = 'black', label=[] , alpha = 0.95, linestyle ='--', linewidth = 1., marker = 's', markersize=3.75):
    """Plot histogram of the scaling exponent, derived from vector data
    """
    vv = aux.computehistogramNoNorm(data,int(bins))
    plt.plot(vv[1],vv[0], color = color, label =label, alpha = alpha, linestyle=linestyle, linewidth = linewidth, marker = marker, markersize=markersize)

def plot_histo_RD(data, bins = 50, color = 'grey', label=[] , alpha = 0.95,  linestyle ='--', linewidth = 1., marker = 'v',markersize=5.):
    """Plot histogram of the scaling exponent, derived from raster data
    """
    vv = aux.computehistogramNoNorm(data.flatten(),int(bins))
    plt.plot(vv[1],vv[0], color = color, label =label, alpha = alpha, linestyle=linestyle, linewidth = linewidth, marker = marker, markersize=markersize)    



"""
======================================================
LomPy (:mod:`lompy`)
======================================================
A python package for the multiscale analysis of spatial point processes

          
.. currentmodule: lompy

======================================================== 
Core functions:

The 'RD' package features yield the multiresolution quantities and the local 
scaling parameters  for uniformly distributed (raster) 2D data

These 'RD' package features yield the multiresolution quantities and local scaling 
parameters for non-uniformly distributed (vector) 2D data

Overview
--------
Routines for the muliresolution quantities:
    
    lompy.MultiresQuantityRD
    lompy.MultiresQuantityVD
    
Routines for the univariate local analysis:
    
    lompy.LocalMsAnalysisRD
    lompy.LocalMsAnalysisVD

Routines for the bivariate local analysis:
    
    lompy.LocalCorrAnalysisRD
    lompy.LocalCorrAnalysisVD

========================================================   
The package includes vizualization routines built on NumPy,
GeoPandas and Matplotlib:

Overview
--------

Vizualization Routine 1:
    
    LomPy.img_arrayVD
    LomPy.img_arrayRD
    
    
Vizualization Routine 2:
    
    LomPy.global_scalingVD
    LomPy.global_scalingRD

Vizualization Routine 3:
    
    LomPy.plot_histo_VD
    LomPy.plot_histo_RD
    
Vizualization Routine 4:
    LomPy.plot_gdf_fit
    LomPy.plot_gdf_radius
    
========================================================   
The package features include a powerful function for linear regression
analysis on large data and other data pre-, and post-processing functions
to aid the workflow with the LomPy package:

Overview
--------
Routines for the linear regression analysis:
    
    lompy.fitwithNan
    
Routines for the histogram computation across scales:
    
    lompy.computehistogramNoNorm

Routines for building a geopandas dataframe from LomPy results:
    
    lompy.build_gdf
    
"""

#Copyright (C) 2023 GNU AGPLv3, LomPy 2023, S.G. Roux and J. Lengyel.
#All rights reserved.
#Contact: stephane.roux@ens-lyon.fr, jankalengyel@gmail.com
#Other Contributors: P. Thiraux, F. Semecurbe

#%%

import warnings
import numpy as np
import pyfftw
import multiprocessing
import types
from sklearn.neighbors import KDTree
from sklearn.neighbors import NearestNeighbors
from functools import partial
import sys
from concurrent.futures import ThreadPoolExecutor
import matplotlib.pyplot as plt
import seaborn as sns
import geopandas as gpd
from scipy.stats import t
import pandas as pd
from shapely.geometry import Polygon


#%%
def checkinputdata(data):
    """
    Check the type of data.
    Image (uniformly sampled)?
    Marked?
    How many marks?
    
    isimage, ismarked, Nmark = checkinputdata(data)

    Parameters
    ----------
    data : numpy array of float
        The data to check.

    Returns
    -------
    isimage : Boolean
        Equal to 1 if the data is an image.
    ismarked : Boolean
        Equal to 1 if the data ia a marked point process.
    Nmark : int
        Number oif mark in the data.

    """
    # check input data
    si=data.shape
    if len(si)==1:
        print('Error : the input argument must have at least two dimensions')
        return
    else:
        
        if (si[0]>100) & (si[1]>100):
            isimage=1
            if len(si)>2:
                #print('The input seems to be a set of  images  of size ({:d},{:d}))'.format(si[2]),si[0])
                print('The input seems to be a set of  {:d}  images  of size ({:d},{:d})).'.format(si[2],si[0],si[1]))
                ismarked = 1
                Nmark = si[2]
            else:
                print('The input seems to be an image of size ({:d},{:})'.format(si[0],si[1]))
                ismarked = 0
                Nmark = 0
        else:
            isimage=0
            if si[1]>2:
                print('The input seems to be a marked point process with {:} points and {:d} marks.'.format(si[0],si[1]-2))
                
                ismarked = 1
                Nmark = si[1]-2
            else:
                print('The input seems to be a marked point process with {:} points.'.format(si[0]))
                ismarked = 0
                Nmark = 0
    return isimage, ismarked, Nmark



"""
================================================================
Auxiliary routines of the LomPy package  (:mod:`lompy.aux`)
================================================================

.. currentmodule: lompy.aux

The package features include a powerful function for linear regression
analysis on large data and other data pre-, and post-processing functions
to aid the workflow with the LomPy package. 


Overview
--------
Routines for the linear regression analysis:
    
    lompy.aux.fitwithNan
    
Routines for the histogram computation across scales:
    
    lompy.aux.computehistogramNoNorm

Routines for building a geopandas dataframe from LomPy results:
    
    lompy.aux.build_gdf

"""

#Copyright (C) 2023 GNU AGPLv3, LomPy 2023, S.G. Roux and J. Lengyel.
#All rights reserved.
#Contact: stephane.roux@ens-lyon.fr, jankalengyel@gmail.com




import numpy as np
import warnings
from scipy.stats import t
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon


#%%
def fitwithNan(Xvector, Yarray, borne=[]):
    '''
        fit a matrix of dim 2 or 3 along first dimension  
        Take care on infinite and nan values in the Yarray
        
        slope, intercept, (pval,rsquare,rmse, stderr) = fitwithNan(Xvector,Yarray,borne=[])
        
        Inputs :
            
            Xvector : vector of x value
        
            Yarray : array od dim 2 or 3 with the size of th first dimension 
                     corresponding to the size of Xvector
                
            borne=[j1, j2] : index of the interval used for then fit  
                           0 < j1 < j2 < len(Xvector). Default j1=0 and j2=len(Xvector).
                         
                          
       Outputs :
           
           slope : the slopes
           intercept : the interceps
           pval : pvalues
           rsquare :  R^2
           rmse     : root mean sqaure errors 
           stderr    : 


     ##

    '''

    if len(borne) == 0:
        j1 = 0
        j2 = Xvector.shape[0]

    else:
        j1 = max(0, borne[0])
        j2 = min(Xvector.shape[0], borne[1])


    if Yarray.ndim == 2:
        Xarraytmp = np.tile(Xvector[:, np.newaxis], (1, Yarray.shape[1]))
        ind = np.isfinite(Yarray)
        Xarray = Xarraytmp*np.nan
        Xarray[ind] = Xarraytmp[ind]
        Xarray = Xarray[j1:j2, :]
        Yarray = Yarray[j1:j2, :]
        lasum = sum(ind[j1:j2, :])

    elif Yarray.ndim == 3:
        Xarraytmp = np.tile(
            Xvector[:, np.newaxis, np.newaxis], (1, Yarray.shape[1], Yarray.shape[2]))
        ind = np.isfinite(Yarray)
        Xarray = Xarraytmp*np.nan
        Xarray[ind] = Xarraytmp[ind]
        Xarray = Xarray[j1:j2, :, :]
        Yarray = Yarray[j1:j2, :, :]
        lasum = sum(ind[j1:j2, :, :])

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)# statistics
        x_mean = np.nanmean(Xarray, axis=0)
        y_mean = np.nanmean(Yarray, axis=0)
        x_std = np.nanstd(Xarray, axis=0)
        y_std = np.nanstd(Yarray, axis=0)
    
        cov = np.nansum((Xarray-x_mean)*(Yarray-y_mean), axis=0)/lasum
        cor = cov/(x_std*y_std)
        slope = cov/(x_std**2)
        intercept = y_mean-x_mean*slope
    
        n = lasum
        tstats = cor*np.sqrt(n-2)/np.sqrt(1-cor**2)
        stderr = slope/tstats
    
        p_val = t.sf(tstats, n-2)*2
        r_square = np.nansum((slope*Xarray+intercept-y_mean) **
                             2, axis=0)/np.nansum((Yarray-y_mean)**2, axis=0)
        rmse = np.sqrt(np.nansum((Yarray-slope*Xarray-intercept)**2, axis=0)/n)

    return slope, intercept, (p_val,r_square,rmse,stderr)  
# %%
def computehistogramNoNorm(WT,bins):
    """
   
      hist, centers, lastd = computehistogram(WT,bins)
      
      Compute  histogram of the normalized wavelet coefficient
      Return also the standart deviation. 
    
      ##
       S.G.  Roux, ENS Lyon, December 2020,  stephane.roux@ens-lyon.fr
       
    """
    if len(WT.shape)==1:
        WT=WT[:,np.newaxis]
        Nr=WT.shape[1]
    else:
        Nr=WT.shape[1]
    hist=np.zeros((bins,Nr))
    centers=np.zeros((bins,Nr))
    for ir in range(Nr):
        temp=WT[:,ir]
        temp=temp[np.isfinite(temp)]

        htmp, bin_edges = np.histogram(temp, bins=bins)
        centers[:,ir]=(bin_edges[:-1]+bin_edges[1:])/2
        dx=np.mean(np.diff(centers[:,ir]))
        hist[:,ir]=htmp/np.sum(htmp)/dx

    return hist, centers #, lastd


# %%
def build_gdf(gridpoints, radius ,results, crs = "EPSG:2154"):
    """
     gdf_results = restoGeoPandaFrame(gridpoints, radius ,results)

         return a geopanda dataframe

     input :

         gridpoints - two dimensional array with the grid points position [x,y]
         radius     - one dimensional array with scales (>0)
         results    - two dimensional array of size equal len(gridpoints) X len(radius)

    output :

        out - geopanda dataframe

    """

    #  grid dataframe
    df_grid = pd.DataFrame({'x':gridpoints[:,0], 'y':gridpoints[:,1]})
    # get all scales in a single dataframe
    j=0
    mystr = 'R'+radius[j].astype(int).astype('str')
    df_data = pd.DataFrame(results[:,j], columns = [mystr])
    for j in range(1,len(radius)):
        mystr = 'R'+radius[j].astype(int).astype('str')
        df_data.loc[:,mystr] = pd.Series(results[:,j], index=df_data.index)

    gridsize = np.abs(df_grid['x'][0] - df_grid['x'][1])
    gdf_results = gpd.GeoDataFrame( df_data, geometry=[Polygon([(x-gridsize/2, y+gridsize/2), (x+gridsize/2, y+gridsize/2), (x+gridsize/2, y-gridsize/2), (x-gridsize/2, y-gridsize/2), (x-gridsize/2, y+gridsize/2)])
                              for x,y in zip(df_grid.x,df_grid.y)])

    gdf_results.crs = crs


    return gdf_results
#%%
def checkinputdata(data):
    """
    Check the type of data.
    Image (uniformly sampled)?
    Marked?
    How many marks?
    
    isimage, ismarked, Nmark = checkinputdata(data)

    Parameters
    ----------
    data : numpy array of float
        The data to check.

    Returns
    -------
    isimage : Boolean
        Equal to 1 if the data is an image.
    ismarked : Boolean
        Equal to 1 if the data ia a marked point process.
    Nmark : int
        Number oif mark in the data.

    """
    # check input data
    si=data.shape
    if len(si)==1:
        print('Error : the input argument must have at least two dimensions')
        return
    else:
        
        if (si[0]>100) & (si[1]>100):
            isimage=1
            if len(si)>2:
                #print('The input seems to be a set of  images  of size ({:d},{:d}))'.format(si[2]),si[0])
                print('The input seems to be a set of  {:d}  images  of size ({:d},{:d})).'.format(si[2],si[0],si[1]))
                ismarked = 1
                Nmark = si[2]
            else:
                print('The input seems to be an image of size ({:d},{:})'.format(si[0],si[1]))
                ismarked = 0
                Nmark = 0
        else:
            isimage=0
            if si[1]>2:
                print('The input seems to be a marked point process with {:} points and {:d} marks.'.format(si[0],si[1]-2))
                
                ismarked = 1
                Nmark = si[1]-2
            else:
                print('The input seems to be a marked point process with {:} points.'.format(si[0]))
                ismarked = 0
                Nmark = 0
    return isimage, ismarked, Nmark


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""

==========================================================
k Nearest Neighbours package for LomPy (:mod:`lompy.knn`)
==========================================================

          
.. currentmodule: lompy.knn

These package features yield the multiresolution quantities and local scaleing parameters 
for non-uniformly distributed (vector) 2D data

Overview
--------
Routines for the muliresolution quantities:
    
    lompy.knn.MultiresQuantityVD
    
Routines for the univariate local analysis:
    
    lompy.knn.LocalMsAnalysisVD

Routines for the bivariate local analysis:
    
    lompy.knn.LocalCorrAnalysisVD

"""
#Copyright (C) 2023 GNU AGPLv3, LomPy 2023, S.G. Roux and J. Lengyel.
#All rights reserved.

#Contact: stephane.roux@ens-lyon.fr, jankalengyel@gmail.com
#Other Contributors: P. Thiraux, F Semecurbe

import warnings
import types
import numpy as np
from sklearn.neighbors import KDTree
from sklearn.neighbors import NearestNeighbors
from functools import partial
import sys
from concurrent.futures import ThreadPoolExecutor
#sys.path.append("../conv2D/") ##!!tag pipchange
#import conv2D as conv2D ##!!tag pipchange
import lompy.conv2D as conv2D

#%%
def checkinputdata(data):
    """
    Check the type of data:
    Image (raster)
    Marked (vector)
    Number of marks
    
    isimage, ismarked, Nmark = checkinputdata(data)

    Parameters
    ----------
    data : numpy array of float
        The data to check.

    Returns
    -------
    isimage : Boolean
        Equals to 1 if the data is an image.
    ismarked : Boolean
        Equals to 1 if the data ia a marked point process.
    Nmark : int
        Number of marks in the data.

    """
    # check input data
    si=data.shape
    if len(si)==1:
        print('Error : the input argument must have at least two dimensions')
        return
    else:
        
        if (si[0]>100) & (si[1]>100):
            isimage=1
            if len(si)>2:
                #print('The input seems to be a set of  images  of size ({:d},{:d}))'.format(si[2]),si[0])
                print('The input is a set of  {:d}  images  of size ({:d},{:d})).'.format(si[2],si[0],si[1]))
                ismarked = 1
                Nmark = si[2]
            else:
                print('The input is an image of size ({:d},{:})'.format(si[0],si[1]))
                ismarked = 0
                Nmark = 0
        else:
            isimage=0
            if si[1]>2:
                print('The input is a marked point process with {:} points and {:d} marks.'.format(si[0],si[1]-2))
                
                ismarked = 1
                Nmark = si[1]-2
            else:
                print('The input is a marked point process with {:} points.'.format(si[0]))
                ismarked = 0
                Nmark = 0
    return isimage, ismarked, Nmark

#%% 
def getoptimblock(datapos, Nanalyse, thresh = 200):
    """
    Define best block positions and sizes.    
    centerstot, sizeblocktot = getoptimblock(datapos, Nanalyse, thresh = 20)
    

    Parameters
    ----------
    datapos : numpy array of float
        list of x and Y position of N points. 
        Array of shape (N, 2).
    Nanalyse : integer
        Number of points used in the batch.
        Must be set accordingly to the memory available.
        
    thresh : float, optional
        Divide block until the number of point inside pass below this threshold.
        The default is 20.

    Returns
    -------
    centerstot : numpy array of float
        List of x and y position of the center of the blocks = shape (Nblock,2).
    sizeblocktot : numpy array of float
        List of size of the blocks = shape (Nblock,).

    """
    
    sizetmp = Nanalyse
    temp = datapos // sizetmp
    block, count = np.unique(temp, axis=0, return_counts = True)
    sizeblock = count * 0 + sizetmp
    centers = block * sizetmp + sizetmp//2   
    
    # above the threshod : we continue to divide
    index, = np.where(count > 2**thresh)
    sizetmp2 = sizetmp // 2
    if len(index)>0:
        counttot  = count[0:index[0]]
        centerstot = centers[0:index[0],:]
        sizeblocktot = sizeblock[0:index[0]]
        
        for iind in range(len(index)):
            iblock = index[iind]
            ii = np.argwhere( np.all((temp - block[iblock]) == 0, axis = 1))
            temp2 = np.squeeze(np.copy(datapos[ii,:]))
            mitemp2 = np.min(temp2,axis = 0)
            temp2 = temp2 - mitemp2[:,np.newaxis].T
            temp2=temp2 // (sizetmp2)
            blocktmp2, counttmp2 = np.unique( temp2, axis = 0, return_counts = True)
            centers2 = blocktmp2 * sizetmp2 + sizetmp2 // 2+ mitemp2
            sizeblock2 = counttmp2 * 0 + sizetmp2
            centerstot = np.append(centerstot,centers2, axis = 0)
            sizeblocktot = np.append(sizeblocktot,sizeblock2, axis = 0)
            counttot = np.append(counttot,counttmp2, axis = 0) 
            #
            if iind + 1 < len(index):
                centerstot = np.append(centerstot, centers[iblock+1:index[iind+1],:], axis = 0)
                sizeblocktot = np.append(sizeblocktot, sizeblock[iblock+1:index[iind+1]], axis = 0)
                counttot = np.append(counttot, count[iblock+1:index[iind+1]], axis = 0)
                
            
        centerstot = np.append( centerstot, centers[index[-1]+1:,:], axis = 0)
        sizeblocktot = np.append( sizeblocktot, sizeblock[index[-1]+1:], axis = 0)
        counttot = np.append( counttot, count[index[-1]+1:], axis = 0)
    else:
        centerstot = centers
        sizeblocktot = sizeblock
        
    return centerstot, sizeblocktot
#
# %%  geo weighting
def EpanechnikovWindow(z):
    return 0.75*(1-z**2) 

def triangularWindow(z):
    return (1 - np.abs(z)) 

def tricubeWindow(z):
    return (70/81)*(1-np.abs(z)**3)**3

def bisquareWindow(z):
    return (1-z**2)**2 

def flatWindow(z):
    return np.ones(z.shape)
#%
def geographicalWeight(dd,T,func):
    
    """
    
    W = geographicalWeight(dist,T,Nr)
    
    return the weight according to the distance dist 
    using a 'local' environment of size T.
    
    Input :

       dist - distance of the point. dist and index as the shame shape
       T    - bandwith of fixed kernel OR bandwith for the pilot density of the adaptive kernel
       Nr   - number of copies for the weight
        
     Output :

       W - the weight. Array of size (len(dd),Nr)
   
    """  

    # Weighting function
    z = dd/T
    W=func(z)
    W[dd>=T]=0
    W=W/np.sum(W,0) 
    W=W[:,np.newaxis]

    return W

#% averaging functions
def localdist(dist,Tloc):
    distout=np.copy(dist)
    distout[dist>Tloc] = np.nan
    distout[dist<=Tloc] = 1
    return distout

#%%
def logaverage(x,w):
    ##########################
    x[x==0] = np.nan
    #w[w==0] = np.nan
    ##########################
    logx=np.log(np.abs(x))
    logx[np.abs(x)==0]=np.nan
    # weight need to be normalized : done
    average=np.nansum(logx*w,axis=0)
    return average

def logaverage2(x,average,w):
    ##########################
    x[x==0] = np.nan
    #w[w==0] = np.nan
    ##########################
    logx=(np.log(np.abs(x))-average)**2
    logx[np.abs(x)==0]=np.nan
    # weight need to be normalized : done
    average2=np.nansum(logx*w,axis=0)
    
    return average2

# %% ---------function for multithreated computations ------ 
def fill(Idx, Dist, val):
    """
    

    Parameters
    ----------
    Idx : TYPE
        DESCRIPTION.
    Dist : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    A : TYPE
        DESCRIPTION.
    B : TYPE
        DESCRIPTION.

    """
    N = len(Idx)
    M = max( map( len, Idx))
    A = np.full( (N, M), np.nan)
    B = np.full( (N, M), np.nan)
                
    #print(A.shape)
    for i, (aa, bb) in enumerate( zip( Idx, Dist)):
        A[i, :len(aa)] = val[aa]
        B[i, :len(bb)] = bb
    return A, B
#
def threaded_Count_Valued(In, radiustot, val):
    """
    

    Parameters
    ----------
    In : TYPE
        DESCRIPTION.
    radiustot : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    Count : TYPE
        DESCRIPTION.

    """
    Idx, Dist = In
    valbis, dd = fill( Idx, Dist, val)
    Count = np.zeros( (valbis.shape[0],len(radiustot),3))
    for ir in range( len( radiustot)):
        
        valbis[dd >= radiustot[-1 - ir]] = np.nan
        Count[:, -1 -ir, 0] = np.nansum( np.isfinite(valbis), axis = 1)
        Count[:, -1 -ir, 1] = np.nanmean( valbis, axis = 1)
        Count[:, -1 -ir, 2] = np.nanstd( valbis, axis = 1)
    
    return Count
#
def fill2(Idx, Dist, val):
    """
    

    Parameters
    ----------
    Idx : TYPE
        DESCRIPTION.
    Dist : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    A : TYPE
        DESCRIPTION.
    B : TYPE
        DESCRIPTION.

    """
    
    N = len(Idx)
    M = max( map( len, Idx))
    A = np.full( (N, M, val.shape[1]), np.nan)
    B = np.full( (N, M), np.nan)
                
    #print(A.shape, val.shape)
    for i, (aa, bb) in enumerate( zip( Idx, Dist)):
        A[i, :len(aa), :] = val[aa,:]
        B[i, :len(bb)] = bb
    return A, B
#
def threaded_All_Valued(In, radiustot, val):
    """
    

    Parameters
    ----------
    In : TYPE
        DESCRIPTION.
    radiustot : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    Count : TYPE
        DESCRIPTION.

    """
    Idx, Dist = In
    valbis, dd = fill2( Idx, Dist, val)
    #print('------', valbis.shape)
    Count = np.zeros( (valbis.shape[0],len(radiustot),7))
    for ir in range( len( radiustot)):
        
        valbis[dd >= radiustot[-1 - ir],-ir] = np.nan
        # count points
        Count[:, -1 -ir, 0] = np.nansum( np.isfinite(valbis[:,:,-ir]), axis = 1)
        # Mean
        Count[:, -1 -ir, 1] = np.nanmean( valbis[:,:,-ir], axis = 1)
        # std
        Count[:, -1 -ir, 2] = np.nanstd( valbis[:,:,-ir], axis = 1)
        # M2
        Count[:, -1 -ir, 3] = np.nanmean( valbis[:,:,-ir]**2, axis = 1)
        # M4
        Count[:, -1 -ir, 4] = np.nanmean( valbis[:,:,-ir]**4, axis = 1)
        # take the log
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            temp  = np.log(np.abs(valbis[:,:,-ir]))           
            temp[np.isinf(temp)] = np.nan
            # c1
            Count[:, -1 -ir, 5] = np.nanmean( temp, axis = 1)
            # c2
            Count[:, -1 -ir, 6] = np.nanstd( temp**2, axis = 1)
    
    return Count


def threaded_corr_coef(In, radiustot, val1,val2):
    """
    

    Parameters
    ----------
    In : TYPE
        DESCRIPTION.
    radiustot : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    Count : TYPE
        DESCRIPTION.

    """
    Idx, Dist = In
    valbis1, dd = fill2( Idx, Dist, val1)
    valbis2, dd = fill2( Idx, Dist, val2)
    corrcoef = np.zeros( (valbis1.shape[0],len(radiustot),2))
    for ir in range( len( radiustot)):
        
        valbis1[dd >= radiustot[-1 - ir],-ir] = np.nan
        valbis2[dd >= radiustot[-1 - ir],-ir] = np.nan
        
        M01 = np.nanmean( valbis1[:,:,-ir], axis = 1)
        M10 = np.nanmean( valbis2[:,:,-ir], axis = 1)
        M11 = np.nanmean( valbis1[:,:,-ir]*valbis2[:,:,-ir], axis = 1)
        S01 = np.nanstd( valbis1[:,:,-ir], axis = 1)
        S10 = np.nanstd( valbis2[:,:,-ir], axis = 1)
    
        corrcoef[:, -1 -ir,0] = (M11-M01*M10)/(S01*S10)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            temp1  = np.log(np.abs(valbis1[:,:,-ir]))           
            temp1[np.isinf(temp1)] = np.nan
            temp2  = np.log(np.abs(valbis2[:,:,-ir]))           
            temp2[np.isinf(temp2)] = np.nan
            M01 = np.nanmean( temp1, axis = 1)
            M10 = np.nanmean( temp2, axis = 1)
            M11 = np.nanmean( temp1*temp2, axis = 1)
            S01 = np.nanstd( temp1, axis = 1)
            S10 = np.nanstd( temp2, axis = 1)
    
            corrcoef[:, -1 -ir,1] = (M11-M01*M10)/(S01*S10)
            
    return corrcoef


#%%
#
def threaded_Log(In, radiustot, val):
    """
    

    Parameters
    ----------
    In : TYPE
        DESCRIPTION.
    radiustot : TYPE
        DESCRIPTION.
    val : TYPE
        DESCRIPTION.

    Returns
    -------
    Count : TYPE
        DESCRIPTION.

    """
    Idx, Dist = In
    Count = np.zeros((len(Idx),len(radiustot)))
    for ir in range(len(radiustot)):
        valbis, dd = fill(Idx,Dist,val[:,ir])
        valbis[dd>=radiustot[-1-ir]]=np.nan
        
        Count[:,-1-ir] = np.nanmean(valbis,axis=1)
    
    return Count

#
def threaded_Log_OneScale(In, radiustot, val):
    Idx, Dist = In
    Count = np.zeros((len(Idx),))
    
    valbis, dd = fill( Idx, Dist, val)
    valbis[dd >= radiustot] = np.nan
    Count = np.nanmean(valbis, axis = 1)
    
    return Count

#
def threaded_Leaders_OneScale(In, radiustot, val,p):
    Idx, Dist = In
    Count = np.zeros((len(Idx),))
    
    valbis, dd = fill( Idx, Dist, val)
    valbis[dd >= radiustot] = np.nan
    Count = np.nansum(np.abs(valbis)**p, axis = 1)**(1/p)
    #Count = Count / np.nansum(np.isfinite(valbis), axis = 1)
    
    return Count

def threaded_MaxLeaders_OneScale(In, radiustot, val):
    Idx, Dist = In
    Count = np.zeros((len(Idx),))
    
    valbis, dd = fill( Idx, Dist, val)
    valbis[dd >= radiustot] = np.nan
    Count = np.nanmax(np.abs(valbis), axis = 1)
    
    return Count

# %% count point with dist<radius
def threaded_Count(dist , radius):   
    """
    
      out = threaded_Count(dist,radius)
      
      Compute  the number of element with distance lower than radius
      Radius can be a vector.
    
       
    """    

    radius = np.atleast_1d(radius)
    Count = np.zeros( ( 1, radius.shape[0]), dtype = 'float')
    for i in range(radius.size):
        dist = dist[ dist< radius[ -1 - i]]
        Count[ 0,-1 - i] = dist.shape[0]
        
    return Count


#%% Leaders knn
def localLeaders(data, Wcoefs, radius, Nanalyse = 2**14, NonUniformData = False, verbose = True, p=2):
    """
    

    Parameters
    ----------
    Wcoef : TYPE
        DESCRIPTION.
    radius : TYPE
        DESCRIPTION.
    Nanalyse : TYPE, optional
        DESCRIPTION. The default is 2**14.
    NonUniformData : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    int
        DESCRIPTION.

    """
    
    #% check input data
    si = data.shape
    if si[1] != 2 : 
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
     
    # select the destination set of points
    destination = data[ :, 0:2]
    N = destination.shape[0]
    
    # check the analysis
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        #lcoefs  = np.log(np.abs(Wcoefs))
        lcoefs  = np.copy(Wcoefs)

        
    Nscales = Wcoefs.shape[1]
    Npts = Wcoefs.shape[0]
    if Npts != N:
        raise TypeError("not a good number of points")
   
    radius = np.atleast_1d(radius)
    radiustot = np.sort(radius)
    temp, temp2, indexreadius = np.intersect1d( radius, radiustot, return_indices = True)
    if radius.shape[0] != Nscales:
        raise TypeError("not a good number of scales")
    # some constants
    scalemax = np.max(radiustot)
    
    
    # sub parameters to partition  the data
    if Nanalyse == 0: # this is dangerous
        print('Warning : dangerous if your set of data is large.')
        Nanalyse = N
        Nbunch = 1
    else:
        # find how many bunches
        if NonUniformData:
            centers, sizeblock = getoptimblock( data, Nanalyse)
            if verbose:
                print('Block optimisation for non uniformly spaced data.')
            Nbunch = len(sizeblock)
        else:
            sizetmp = Nanalyse
            temp=data // sizetmp
            block, count = np.unique( temp, axis = 0, return_counts = True)
            centers = block*sizetmp + sizetmp // 2
            sizeblock = count * 0 + sizetmp
            Nbunch = len(block)
            if verbose:
                print('No block optimisation uniformly spaced data.')
            
    
    Nanalyse2 = 2**22
    
    if verbose:
        print('Computation in {:d} bunches :'.format(Nbunch), end = ' ')
    
    # allocation
    Lead = np.nan*np.zeros((N, radius.shape[0]), dtype = float)
    LeadMax = np.nan*np.zeros((N, radius.shape[0]), dtype = float)

    
    # set the worker for knn search
    neigh = NearestNeighbors( n_jobs = 4)
    
    #% loop on the bunches
    for ibunch in range( Nbunch):
        #%
        print(ibunch + 1, end=' ')
        sys.stdout.flush()
              
        center = centers[ ibunch, :]
        sizetmp = sizeblock[ ibunch]
        #grrr
        index,= np.where(((data[:,0] >= center[0]-sizetmp//2) & (data[:,0]< center[0]+sizetmp//2)) & ((data[:,1]>= center[1]-sizetmp//2) & (data[:,1]< center[1]+sizetmp//2)))
        IndexNear, = np.where(((data[:,0] >= center[0]-(sizetmp//2+scalemax+1)) & (data[:,0]< center[0]+(sizetmp//2+scalemax+1))) & ((data[:,1]>= center[1]-(sizetmp//2+scalemax+1)) & (data[:,1]< center[1]+(sizetmp//2+scalemax+1))))
        neigh.fit( data[ IndexNear, :])
        
        # search neiarest neighbors
        Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2], radius = scalemax, return_distance = True)
        # get the max length of neighbors --> computation by block
        Maxlength = max(map(len, IdxTot))
        Nblock = int(np.ceil( index.shape[0] * Maxlength / Nanalyse2))
        Nptx = len(index) // Nblock+1
        # correct the number of blocks if needed
        if (Nblock -1) * Nptx > min( len(index), Nblock*Nptx):
            Nblock = Nblock - 1
        
        
        for ir in range( Nscales):
            radius = radiustot[ ir]
            malist2 = [(IdxTot[i*Nptx:min(len(index),(i+1)*Nptx)], Disttot[i*Nptx:min(len(index),(i+1)*Nptx)]) for i in range(Nblock)]
           
            partialknnMeanLog = partial( threaded_Leaders_OneScale, radiustot = radius, val = lcoefs[IndexNear,ir], p = p)
            with ThreadPoolExecutor() as executor:
                result_list4 = executor.map( partialknnMeanLog, malist2)
            
            Lead[ index, ir] = np.hstack( list( result_list4))
            partialknnMeanLog = partial( threaded_MaxLeaders_OneScale, radiustot = radius, val = lcoefs[IndexNear,ir])
            with ThreadPoolExecutor() as executor:
                result_list4 = executor.map( partialknnMeanLog, malist2)
            
            LeadMax[ index, ir] = np.hstack( list( result_list4))
            
    
    for ir in range(1, Nscales):
        LeadMax[ index, ir] = np.maximum(LeadMax[:, ir], LeadMax[:, ir-1])
        
    print('.')    
    return Lead, LeadMax

#%% Leaders knn
def localLeadersP(data, Wcoefs, radius, Nanalyse = 2**14, NonUniformData = False, verbose = True, p=2):
    """
    

    Parameters
    ----------
    Wcoef : TYPE
        DESCRIPTION.
    radius : TYPE
        DESCRIPTION.
    Nanalyse : TYPE, optional
        DESCRIPTION. The default is 2**14.
    NonUniformData : TYPE, optional
        DESCRIPTION. The default is False.

    Returns
    -------
    int
        DESCRIPTION.

    """
    
    #% check input data
    si = data.shape
    if si[1] != 2 : 
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
     
    # select the destination set of points
    destination = data[ :, 0:2]
    N = destination.shape[0]
    
    # check the analysis
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        #lcoefs  = np.log(np.abs(Wcoefs))
        lcoefs  = np.copy(Wcoefs)

    Nscales = Wcoefs.shape[1]
    Npts = Wcoefs.shape[0]
    if Npts != N:
        raise TypeError("not a good number of points")
   
    radius = np.atleast_1d(radius)
    radiustot = np.sort(radius)
    temp, temp2, indexreadius = np.intersect1d( radius, radiustot, return_indices = True)
    if radius.shape[0] != Nscales:
        raise TypeError("not a good number of scales")
    # some constants
    scalemax = np.max(radiustot)
    
    
    # sub parameters to partition  the data
    if Nanalyse == 0: # this is dangerous
        print('Warning : dangerous if your set of data is large.')
        Nanalyse = N
        Nbunch = 1
    else:
        # find how many bunches
        if NonUniformData:
            centers, sizeblock = getoptimblock( data, Nanalyse)
            if verbose:
                print('Block optimisation for non uniformly spaced data.')
            Nbunch = len(sizeblock)
        else:
            sizetmp = Nanalyse
            temp=data // sizetmp
            block, count = np.unique( temp, axis = 0, return_counts = True)
            centers = block*sizetmp + sizetmp // 2
            sizeblock = count * 0 + sizetmp
            Nbunch = len(block)
            if verbose:
                print('No block optimisation uniformly spaced data.')
            
    
    Nanalyse2 = 2**22
    
    if verbose:
        print('Computation in {:d} bunches :'.format(Nbunch), end = ' ')
    
    # allocation
    Lead = np.nan*np.zeros((N, radius.shape[0]), dtype = float)

    
    # set the worker for knn search
    neigh = NearestNeighbors( n_jobs = 4)
    
    #% loop on the bunches
    for ibunch in range( Nbunch):
        #%
        print(ibunch + 1, end=' ')
        sys.stdout.flush()
              
        center = centers[ ibunch, :]
        sizetmp = sizeblock[ ibunch]
        #grrr
        index,= np.where(((data[:,0] >= center[0]-sizetmp//2) & (data[:,0]< center[0]+sizetmp//2)) & ((data[:,1]>= center[1]-sizetmp//2) & (data[:,1]< center[1]+sizetmp//2)))
        IndexNear, = np.where(((data[:,0] >= center[0]-(sizetmp//2+scalemax+1)) & (data[:,0]< center[0]+(sizetmp//2+scalemax+1))) & ((data[:,1]>= center[1]-(sizetmp//2+scalemax+1)) & (data[:,1]< center[1]+(sizetmp//2+scalemax+1))))
        neigh.fit( data[ IndexNear, :])
        
        # search neiarest neighbors
        Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2], radius = scalemax, return_distance = True)
        # get the max length of neighbors --> computation by block
        Maxlength = max(map(len, IdxTot))
        Nblock = int(np.ceil( index.shape[0] * Maxlength / Nanalyse2))
        Nptx = len(index) // Nblock+1
        # correct the number of blocks if needed
        if (Nblock -1) * Nptx > min( len(index), Nblock*Nptx):
            Nblock = Nblock - 1
        
        
        for ir in range( Nscales):
            radius = radiustot[ ir]
            malist2 = [(IdxTot[i*Nptx:min(len(index),(i+1)*Nptx)], Disttot[i*Nptx:min(len(index),(i+1)*Nptx)]) for i in range(Nblock)]
           
            partialknnMeanLog = partial( threaded_Leaders_OneScale, radiustot = radius, val = lcoefs[IndexNear,ir], p = p)
            with ThreadPoolExecutor() as executor:
                result_list4 = executor.map( partialknnMeanLog, malist2)
            
            Lead[ index, ir] = np.hstack( list( result_list4))

        
    print('.')    
    return Lead
#%%
def localWaveTrans(data, radius, Nanalyse = 2**14, destination = np.array([]),NonUniformData = False, verbose = True):
    """
    
     [Count, Wave, CountG] = aveTrans(source, radius, T=None, Nanalyse=2**16, destination = []))
    
    
    
    Compute box-counting and wavelet coefficient on a valued/non valued set of data points.
    
    If the data are not valued, count for every data point
            -- the number of neighboors in ball of radius r (GWFA) : N(r).
            -- the wavelet coeficient at scale r (GWMFA) : 2*N(r)-N(sqrt(2)*r).
    
    If the data are valued, count for every datapoint
            -- the number of neighboors in ball of radius r, the mean and std of the marked value
            -- the wavelet coeficient at scale r on the marked value.
    
    Input :
    
        source     - Non-marked  point process : matrix of size N X 2 for N points
                        where source[i,:]=[X_i,Y_i] with 2D cooprdonate of point i.
                      Marked  point process :  matrix of size Nx3
                        where source[i,:]=[X_i,Y_i, mark_i] with 2D cooprdonate of point i and value
        radius      - list of scales to be investigated
        Nanalyse    - number of points to analyse in one bach. Default is 2**16 points.
                        If Nanalyse=0, compute all the points in once (dangerous!!!)
        destination - Non marked point process (destination_i=[X_I,Y_i]) where the coeficient are calculated
                        Default empty : compute at source position.
    Output :
    
        Count    - matrix of size Nxlength(radius) with box-counting coefficients
        Wave     - matrix of size Nxlength(radius) with wavelet coefficients
        Count    - matrix of size Nx2 with box-counting and wavelet coeficient at scale T
                  
    
    """
 
    #% check input data
    si = data.shape
    if si[1] < 2: # or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
     
    
    
    # select the destination set of points
    if destination.shape[0] == 0: 
        destination = data[:,0:2]
    else:
        sid = destination.shape
        #print(sid)
        if sid[1] != 2 :
            raise TypeError('The ''destination''  argument must be of length 2.')
    
    # check the analysis
    isvalued = 0 
    if si[1] > 2: # valued analysis
        isvalued = 1
        if verbose:
            print('Valued data analysis on {:d} points to {:d} destination points.'.format(data.shape[0],destination.shape[0]))
        val = data[:,2]
        #valtot = data[:,2:]
        #data = data[:,0:2]
        
    else:           # non valued analysis
        val = destination[:,0] * 0 + 1
        if verbose:
            print('Non valued data analysis on {:d} points to {:d} destination points.'.format(data.shape[0],destination.shape[0]))
    
    radius = np.atleast_1d(radius)
    radiustot = np.sort(radius)
    #temp,temp2,indexreadius=np.intersect1d(radius,radiustot, return_indices=True)

    # some constants
    scalemax=np.max(radiustot)
    N = destination.shape[0]
    
    # sub parameters to partition  the data
    if Nanalyse==0: # this is dangerous
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=N
        Nbunch=1
    else:
        # find how many bunchs
        if NonUniformData:
            centers, sizeblock = getoptimblock(destination,Nanalyse)
            Nbunch = len(sizeblock)
            if verbose:
                print('Block optimisation for non uniformly spaced data.')
                       
        else:
            sizetmp = Nanalyse
            temp= destination // sizetmp
            block, count = np.unique(temp, axis=0, return_counts=True)
            centers = block*sizetmp+sizetmp//2
            sizeblock = count*0+sizetmp
            Nbunch = len(block)
            if verbose:
                print('No block optimisation uniformly spaced data.')
            
    
    Nanalyse2 = 2**22
    #print('Nanalyse (bunch) and Nanalyse2 (block)',  Nanalyse, Nanalyse2,t2-t1)
    if verbose:
        print('Computation in {:d} bunches :'.format(Nbunch),end=' ')  
    
   
    if isvalued: # we compute mean, std and count
        Count = np.nan*np.zeros((N,radius.shape[0],3),dtype=float)
        #  store count of valid value, mean and std of the point process mark           
    else:         #  we compute only count
        Count = np.nan*np.zeros((N,radius.shape[0]),dtype=float)
    
    # set the worker for knn search
    neigh = NearestNeighbors(n_jobs=4)
    
    #% loop on the bunches
    for ibunch in range(Nbunch):
        #%
        #ibunch=4
        print(ibunch+1, end=' ')
        sys.stdout.flush()
        #
        center = centers[ibunch,:]
        sizetmp = sizeblock[ibunch]
        #
        index, = np.where(((destination[:,0] >= center[0]-sizetmp//2) & (destination[:,0]< center[0]+sizetmp//2)) & ((destination[:,1]>= center[1]-sizetmp//2) & (destination[:,1]< center[1]+sizetmp//2)))
        IndexNear, = np.where(((destination[:,0] >= center[0]-(sizetmp//2+scalemax+1)) & (destination[:,0]< center[0]+(sizetmp//2+scalemax+1))) & ((destination[:,1]>= center[1]-(sizetmp//2+scalemax+1)) & (destination[:,1]< center[1]+(sizetmp//2+scalemax+1))))
        neigh.fit(destination[IndexNear,:])
        
        if isvalued:  # valued analysis

            Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2], radius = scalemax, return_distance = True)

            Maxlength = max(map( len, IdxTot))
            
            Nblock = int(np.ceil(index.shape[0] * Maxlength / Nanalyse2))           
            Nptx = len(index) // Nblock + 1
            if (Nblock - 1)*Nptx > min(len(index), Nblock * Nptx):
                Nblock = Nblock - 1
            
            #print(Nblock,Nptx)
            malist2=[(IdxTot[i*Nptx:min(len(index),(i+1)*Nptx)], Disttot[i*Nptx:min(len(index),(i+1)*Nptx)]) for i in range(Nblock)]
            partialknncountbis = partial(threaded_Count_Valued, radiustot = radiustot,val = val[IndexNear])

            
            with ThreadPoolExecutor() as executor:
                result_list3 = executor.map(partialknncountbis, malist2)
            
            Count[index,:,:] = np.vstack(list(result_list3))

            
        else:  # non valued analysis
            Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2],radius=scalemax,return_distance=True)
            partialcount = partial(threaded_Count, radius=radiustot)
            with ThreadPoolExecutor() as executor:
                result_list = executor.map(partialcount, Disttot)
            
            temp = np.squeeze(np.array(list(result_list)));
            if radius.shape[0] == 1:
                Count[index,:] = temp[:,np.newaxis]
            else:    
                Count[index,:] = temp
            
    del neigh
    print('.')
    
    if isvalued: # compute poor coefs       
        Wcoef = np.copy(Count)
        Wcoef[:,:,1] = Wcoef[:,:,1] - val[:,np.newaxis]    
        return Wcoef, Count
    
    else:
        return Count

# %%
def computeGridValue(tree,Count,lCount,IndexNear,index2,mygrid, radius, Tloc, weightingfunction):
    """

    NWmean, NWstd, Wmean, Wstd, Mom, Cum = computeGridValue(tree,Count,IndexNear,index,mygrid, radius, T, k=0):

    For marked and non-marked point process
    Compute weighted statistics of the number of point location or value.
    The (geographical) weighting is obtained using the bi-squared function of size T.
       
    Input :

       tree      - a kdtree obtained on the IndexNear points (a subset of the Npts total points)
       Count     - matrix of size Nptsxlength(radius) with wavelet coef
       IndexNear - location of the points used for the kdtree
       index     - index of the grid points under investigations
       mygrid    - all the grid points (size (Ngridpts, 2) with X and Y spatial location
       radius   - list of scales to be investigated
       T        - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       k        - number of min neighbors for the adaptive kernel smoothing. If k=0 : fixed kernel.
        

     Output :

       res - list of two dimensional numpy matrix (of size length(data) X length(radius)

              NWmean  : non weighted mean
              NWstd   : non weighted standart deviation
              Wmean   : weighted mean
              Wstd    : weighted standart deviation
              Mom     : weighted moment (order 0 to 4) of the absolute value of the coefficient. 
              Cum     : weighted cumulant (order one and two) of the coefficient. 


    """

    # make the tree accordingly to k
    Tmax = np.nanmax(Tloc)
    
    neighbors_i_fixed, dist_ie_fixed = tree.query_radius(mygrid[index2,:], r = Tmax, count_only=False, return_distance=True, sort_results=False)

    
    print('.',end='')
    sys.stdout.flush()
 
    # non weighted coef
    tmp_fixed2=[ Count[IndexNear[neighbors_i_fixed[igrid]],:] for igrid in range (len(index2))]
    lead_tmp_fixed2=[ lCount[IndexNear[neighbors_i_fixed[igrid]],:] for igrid in range (len(index2))]
    # number of  coef
    ltmp=np.array([ len(dist_ie_fixed[igrid]) for igrid in range (len(index2))])
    # the weight
    Wfinal2=[  weightingfunction(dist_ie_fixed[igrid], Tloc[igrid]) for igrid in range (len(index2))]
    # compute D0 and moments
    D0 = np.array([(np.nansum( Wfinal2[igrid]*np.abs(tmp_fixed2[igrid])**-1  , axis=0)) for igrid in range(len(index2))])
    D0 = D0**-1
    print('.',end='')
    Mom1 = np.array([np.nansum( np.abs(tmp_fixed2[igrid]) * Wfinal2[igrid], axis=0)  for igrid in range(len(index2))])
    print('.',end='')
    Mom2 = np.array([np.nansum( tmp_fixed2[igrid]**2 * Wfinal2[igrid], axis=0)  for igrid in range(len(index2))])
    print('.',end='')
    print('.',end='')
    Mom4 = np.array([np.nansum( tmp_fixed2[igrid]**4 * Wfinal2[igrid], axis=0)  for igrid in range(len(index2))])
    print('. ',end='')
    sys.stdout.flush()
    
    # compute Non weighted std
    newdist=[ localdist(dist_ie_fixed[igrid],Tloc[igrid])  for igrid in range(len(index2))]
   
    NWstd = np.array([np.nanstd( tmp_fixed2[igrid]*newdist[igrid][:,np.newaxis], axis=0, ddof=1) for igrid in range(len(index2))])
    NWstd[ltmp<=10,:] =0
    NWmean = np.array([np.nanmean( tmp_fixed2[igrid]*newdist[igrid][:,np.newaxis], axis=0) for igrid in range(len(index2))])
    NWmean[ltmp<=10,:]=0          
    print('*',end='')
    sys.stdout.flush()
    
    # compute weighted mean and std
    average = np.array([ np.nansum(tmp_fixed2[igrid]*Wfinal2[igrid], axis=0) for igrid in range(len(index2))])
    Wmean = average
    Wmean[ltmp<=10,:]=0      
    Wstd  = np.array([np.nansum((tmp_fixed2[igrid]-average[igrid])**2*Wfinal2[igrid], axis=0) for igrid in range(len(index2))])
    Wstd[ltmp<=10,:]=0
    Wstd = np.sqrt(Wstd)
    print('* ',end='')
    
    # compute Weighted cumulant
    average= np.array([logaverage(tmp_fixed2[igrid], Wfinal2[igrid]) for igrid in range(len(index2))])    
    Cum1 = average
    Cum1[ltmp<=10,:]=0     
    print('+',end='')
    lead_average =  np.array([logaverage(lead_tmp_fixed2[igrid], Wfinal2[igrid]) for igrid in range(len(index2))])
    Cum2=np.array([logaverage2(lead_tmp_fixed2[igrid],lead_average[igrid], Wfinal2[igrid]) for igrid in range(len(index2))])
    Cum2[ltmp<=10,:]=0   
    print('+ ',end='')
    
    Mean_std=np.stack([Wmean, Wstd], axis = 0)
    Mom12=np.stack([ Mom1, Mom2],axis=0)
    Cum=np.stack([Cum1, Cum2],axis = 0)
    Flatness=Mom4/(3*Mom2**2)
    sys.stdout.flush()   

    return Mean_std, Mom12, Flatness, Cum
# %% Local Multiscale Analysis
# add Tloc and function
def LocalMsAnalysisVD(data,Count,lCount,X,Y,radius, T, adaptive = False, weights = 'flat',  Nanalyse=2**16, Tloc=[]):
    """

    res = WaveSmoothingOptim(data,Wave,X,Y,radius,T,Nanalyse=2**16, k = 0, isvalued=0))

    For marked and non-marked point process
    Compute kernel smoothing of the wavelet coefficient of a dataset of points.
    The geographical weighting is obtained using the bi-squared function of size T.
   
    Input :

       data     - matrix of size Nx2 --> position (x,y) for N points
       mq       - matrix of size Nxlength(radius) with multiresolution quantity count
                  Can be obtained using the function  LomPy.m
       X        - array of dim 2 with x-postion of the grid nodes
       Y        - array of dim 2 with y-postion of the grid nodes : X and Y must have the same size
       radius   - list of scales to be investigated
       T        - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       Nanalyse - number of points to analyse in one bach. Default is 2**16 points.
       G        - mean of all pilot density estimates
       isvalued - boolean. If = 1 remove the grid point where the value of the point inside thegrid pixel are all 0.
       

     Output :

       res - list of two dimensional numpy matrix (of size length(data) X length(radius)

              res[Ø] = Wmean   : weighted mean
              res[1] = Wstd    : weighted standart deviation
              res[2] = NWratio : non weighted mean
              res[3] = NWstd   : non weighted standart deviation
              res[4] = Mom     : weighted moment (order 0 to 4) of the absolute value
                      of thecoefficient. matrix of size 5 X length(data) X length(radius)
              res[5] = Cum     : weighted cumulant one and two 
                      of thecoefficient. matrix of size 2 X length(data) X length(radius)
    """


    si=data.shape
    if si[1] < 2 or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
        
    if Count.ndim > 2:
        raise TypeError('The second arument must be two or threedimensional.')

    if Count.shape[0] != si[0]:
        raise TypeError('The twofirst arument must have the same length.')

    if radius[radius>T].shape[0] >0:
        raise TypeError('The last argument must be greater than a sqrt(2)*radius.')

    # check grid
    if np.sum(np.abs(np.array(X.shape)-np.array(Y.shape)))>0:
        raise TypeError('X and Y must have the same size.')

    if np.sum(np.diff(X,axis=1,n=2))>0:
        raise TypeError('X must be regulary sampled.')
    else:
        gridsizex=max(X[0][1] - X[0][0],X[1][0] - X[0][0])
        
    if np.sum(np.diff(Y,axis=0,n=2))>0:
        raise TypeError('Y must be regulary sampled.')
    else:
        gridsizey=max(Y[1][0] - Y[0][0],Y[0][1] - Y[0][0])
    
    if gridsizex != gridsizey:
        raise TypeError('X and Y must have same sampling.')
    
    if radius.size != Count.shape[1]:
        raise TypeError('The size of the second arument must be [N,R,M] where R is the number of scales (=radius.size)')
     
    
    # create grid points
    mygrid=np.column_stack([X.flatten(),Y.flatten()])
    gridlength=mygrid.shape[0]
    gridsizecurr = gridsizex
    
    Nanalyse2=300000
    if Nanalyse==0:
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=gridlength
        Nbunch=1
    else: # find how many bunchs       
        Nbunch = np.ceil(gridlength/Nanalyse).astype(int)

    if isinstance(weights, str):
        if weights == 'flat':
            weightingfunction = partial(geographicalWeight,func = flatWindow)
            print('No weight')
        elif weights == 'Epanechnikov':
            weightingfunction = partial(geographicalWeight,func = EpanechnikovWindow)
        elif weights == 'bisquare':
            weightingfunction = partial(geographicalWeight,func = bisquareWindow)
        elif weights == 'triangular':
            weightingfunction = partial(geographicalWeight,func = triangularWindow)
        elif weights == 'tricube':
            weightingfunction = partial(geographicalWeight,func = tricubeWindow)
            
    elif isinstance(weights, types.FunctionType):
        weightingfunction = partial(geographicalWeight,func = weights)
    else:
        raise TypeError('ERROR : weighting is of unknown type.).')
        
    
    # results allocations    
    Cum = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    Mean_Std = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    Mom12 = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    Flatness = np.nan*np.zeros( (gridlength,radius.size), dtype=float)
    if adaptive == True:
        if len(Tloc) ==0: # compute local env
            print('ADAPTIVE KERNEL: Computing pilot density')
            CountT= localWaveTrans(data[:,(0,1)], np.atleast_1d(T))
            WmeanStd_T, Mom_T, Flat_T, Cum_T = LocalMsAnalysisVD(data[:,(0,1)],CountT,X,Y,np.atleast_1d(T), T=T,weights = 'Epanechnikov',Nanalyse=2**16)
            Wmean_Tloc = WmeanStd_T[0,:,0]
            G = np.nanmean(Wmean_Tloc)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                lambda_ie = G/Wmean_Tloc
            Tloc = lambda_ie*T
            Tloc[np.isinf(Tloc)] = T
            Tlocmax = np.nanquantile(Tloc,0.95)#.9
            Tloc[Tloc>Tlocmax] = Tlocmax 
            Tlocmin = np.nanquantile(Tloc,0.05)#.05
            Tloc[Tloc<Tlocmin] = Tlocmin 
            T=Tlocmax
            print('ADAPTIVE KERNEL computed.')
        elif X.flatten().shape[0] == Tloc.flatten().shape[0]:
            T=np.nanmax(Tloc)
        else: 
            raise TypeError('ERROR : wrong size for the local environment')
    else:
        Tloc=T*np.ones(X.size)
        
    # Loop on bunch
    for ibunch in range(Nbunch):
        # %
        print('bunch {:d}/{:d} '.format(ibunch+1,Nbunch), end=' ')
        sys.stdout.flush()
        # get data index of current bunch
        index=np.arange(ibunch*Nanalyse,(ibunch+1)*Nanalyse,1)
        index=index[index<gridlength]

        # we restrict the tree to points whithin a radius T (which must be > radius2)
        mi=np.min(mygrid[index,:], axis=0)
        ma=np.max(mygrid[index,:], axis=0)
        IndexNear=np.where((data[:,0] >mi[0]-T) & (data[:,0] <ma[0]+T) & (data[:,1]  >mi[1]-T) & (data[:,1] <ma[1]+T))[0]
         
        # make the tree with the nearest points only
        if IndexNear.shape[0]>0:
                #print('coucou',len(IndexNear))
                tree = KDTree(data[IndexNear,0:2])
                Idxtot = tree.query_radius(mygrid[index,:], r =  np.sqrt(2)*(gridsizecurr/2), count_only=False, return_distance=False)

                # find and remove empty (no point location) grid pixels
                thelengths=np.array([ Idxtot[igrid].shape[0] for igrid in range (len(Idxtot))])
                IdxMin, = np.where(thelengths>0.)
                index = index[IdxMin]
                Idxtot=Idxtot[IdxMin]
                thelengths=thelengths[IdxMin]
                    
                # Managed the number of grid points  (too many neighboors for some)
                cumsumbunch=np.cumsum(thelengths)
                
                Nflowers=int(np.ceil(np.sum(thelengths)/Nanalyse2))
                print('{:d} batch(s).'.format(Nflowers))
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    
                    for iflower in range(Nflowers):
                        i1,=np.where(iflower*Nanalyse2 <= cumsumbunch)
                        i2,=np.where(cumsumbunch <  (iflower+1)*Nanalyse2)
                        flowers=np.intersect1d(i1,i2)
                        Mean_Std[:,index[flowers],:], Mom12[:,index[flowers],:], Flatness[index[flowers],:],Cum[:,index[flowers],:]=computeGridValue(tree,Count,lCount,IndexNear,index[flowers],mygrid, radius,Tloc[index[flowers]],weightingfunction)
                
                print('')

    if adaptive == True:
        return Mean_Std, Mom12, Flatness, Cum, Tloc
    else:
        return Mean_Std, Mom12, Flatness, Cum
        
    
# %% Local Multiscale Analysis
# add Tloc and function
def LmsAnalysisPP(data,Count,lCount,X,Y,radius, T, adaptive = False, weights = 'flat',  Nanalyse=2**16, Tloc=[]):
    """

    res = WaveSmoothingOptim(data,Wave,X,Y,radius,T,Nanalyse=2**16, k = 0, isvalued=0))

    For marked and non-marked point process
    Compute kernel smoothing of the wavelet coefficient of a dataset of points.
    The geographical weighting is obtained using the bi-squared function of size T.
   
    Input :

       data     - matrix of size Nx2 --> position (x,y) for N points
       Wave     - matrix of size Nxlength(radius) with wavelet count
                  Can be obtained using the function  GWFA_count.m
       X        - array of dim 2 with x-postion of the grid nodes
       Y        - array of dim 2 with y-postion of the grid nodes : X and Y must have the same size
       radius   - list of scales to be investigated
       T        - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       Nanalyse - number of points to analyse in one bach. Default is 2**16 points.
       G        - mean of all pilot density estimates
       isvalued - boolean. If = 1 remove the grid point where the value of the point inside thegrid pixel are all 0.
       

     Output :

       res - list of two dimensional numpy matrix (of size length(data) X length(radius)

              res[Ø] = Wmean   : weighted mean
              res[1] = Wstd    : weighted standart deviation
              res[2] = NWratio : non weighted mean
              res[3] = NWstd   : non weighted standart deviation
              res[4] = Mom     : weighted moment (order 0 to 4) of the absolute value
                      of thecoefficient. matrix of size 5 X length(data) X length(radius)
              res[5] = Cum     : weighted cumulant one and two 
                      of thecoefficient. matrix of size 2 X length(data) X length(radius)

    """


    si=data.shape
    if si[1] < 2 or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
        
    if Count.ndim > 2:
        raise TypeError('The second arument must be two or threedimensional.')

    if Count.shape[0] != si[0]:
        raise TypeError('The twofirst arument must have the same length.')

    if radius[radius>T].shape[0] >0:
        raise TypeError('The last argument must be greater than a sqrt(2)*radius.')

    # # check grid
    # if np.sum(np.abs(np.array(X.shape)-np.array(Y.shape)))>0:
    #     raise TypeError('X and Y must have the same size.')

    # if np.sum(np.diff(X,axis=1,n=2))>0:
    #     raise TypeError('X must be regulary sampled.')
    # else:
    #     gridsizex=max(X[0][1] - X[0][0],X[1][0] - X[0][0])
        
    # if np.sum(np.diff(Y,axis=0,n=2))>0:
    #     raise TypeError('Y must be regulary sampled.')
    # else:
    #     gridsizey=max(Y[1][0] - Y[0][0],Y[0][1] - Y[0][0])
    
    # if gridsizex != gridsizey:
    #     raise TypeError('X and Y must have same sampling.')
    
    if radius.size != Count.shape[1]:
        raise TypeError('The size of the second arument must be [N,R,M] where R is the number of scales (=radius.size)')
     
    
    # create grid points
    mygrid=np.column_stack([X.flatten(),Y.flatten()])
    gridlength=mygrid.shape[0]
    #gridsizecurr = gridsizex
    
    Nanalyse2=300000
    if Nanalyse==0:
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=gridlength
        Nbunch=1
    else: # find how many bunchs       
        Nbunch = np.ceil(gridlength/Nanalyse).astype(int)
    
    #print('Computing in {:d} bunch(es)'.format(Nbunch))
    # choose weighting
    
    if isinstance(weights, str):
        if weights == 'flat':
            weightingfunction = partial(geographicalWeight,func = flatWindow)
            print('No weight')
        elif weights == 'Epanechnikov':
            weightingfunction = partial(geographicalWeight,func = EpanechnikovWindow)
        elif weights == 'bisquare':
            weightingfunction = partial(geographicalWeight,func = bisquareWindow)
        elif weights == 'triangular':
            weightingfunction = partial(geographicalWeight,func = triangularWindow)
        elif weights == 'tricube':
            weightingfunction = partial(geographicalWeight,func = tricubeWindow)
            
    elif isinstance(weights, types.FunctionType):
        weightingfunction = partial(geographicalWeight,func = weights)
    else:
        raise TypeError('ERROR : weighting is of unknown type.).')
        
    
    # results allocations    
    Cum = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    # new alloc :
    Mean_Std = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    Mom12 = np.nan*np.zeros( (2, gridlength,radius.size), dtype=float)
    Flatness = np.nan*np.zeros( (gridlength,radius.size), dtype=float)
    if adaptive == True:
        if len(Tloc) ==0: # compute local env
            print('ADAPTIVE KERNEL: Computing pilot density')
            CountT= localWaveTrans(data[:,(0,1)], np.atleast_1d(T))
            WmeanStd_T, Mom_T, Flat_T, Cum_T = LocalMsAnalysisVD(data[:,(0,1)],CountT,X,Y,np.atleast_1d(T), T=T,weights = 'Epanechnikov',Nanalyse=2**16)
            Wmean_Tloc = WmeanStd_T[0,:,0]
            G = np.nanmean(Wmean_Tloc)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                lambda_ie = G/Wmean_Tloc
            Tloc = lambda_ie*T
            Tloc[np.isinf(Tloc)] = T
            Tlocmax = np.nanquantile(Tloc,0.95)#.9
            Tloc[Tloc>Tlocmax] = Tlocmax 
            Tlocmin = np.nanquantile(Tloc,0.05)#.05
            Tloc[Tloc<Tlocmin] = Tlocmin 
            T=Tlocmax
            print('ADAPTIVE KERNEL computed.')
        elif X.flatten().shape[0] == Tloc.flatten().shape[0]:
            T=np.nanmax(Tloc)
        else: 
            raise TypeError('ERROR : wrong size for the local environment')
    else:
        Tloc=T*np.ones(X.size)
        
    # Loop on bunch
    for ibunch in range(Nbunch):
        # %
        print('bunch {:d}/{:d} '.format(ibunch+1,Nbunch), end=' ')
        sys.stdout.flush()
        # get data index of current bunch
        index=np.arange(ibunch*Nanalyse,(ibunch+1)*Nanalyse,1)
        index=index[index<gridlength]

        # we restrict the tree to points whithin a radius T (which must be > radius2)
        mi=np.min(mygrid[index,:], axis=0)
        ma=np.max(mygrid[index,:], axis=0)
        IndexNear=np.where((data[:,0] >mi[0]-T) & (data[:,0] <ma[0]+T) & (data[:,1]  >mi[1]-T) & (data[:,1] <ma[1]+T))[0]
         
        # make the tree with the nearest points only
        if IndexNear.shape[0]>0:
                tree = KDTree(data[IndexNear,0:2])
                Idxtot = tree.query_radius(mygrid[index,:], r = np.nanmax(radius), count_only=False, return_distance=False)
                # find and remove empty (no point location) grid pixels
                thelengths=np.array([ Idxtot[igrid].shape[0] for igrid in range (len(Idxtot))])
                IdxMin, = np.where(thelengths>0.)
                index = index[IdxMin]
                Idxtot=Idxtot[IdxMin]
                thelengths=thelengths[IdxMin]
            
                cumsumbunch=np.cumsum(thelengths)
                
                Nflowers=int(np.ceil(np.sum(thelengths)/Nanalyse2))
                print('{:d} batch(s).'.format(Nflowers))
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    
                    #start = time.process_time()
                    for iflower in range(Nflowers):
                        i1,=np.where(iflower*Nanalyse2 <= cumsumbunch)
                        i2,=np.where(cumsumbunch <  (iflower+1)*Nanalyse2)
                        flowers=np.intersect1d(i1,i2)
                        Mean_Std[:,index[flowers],:], Mom12[:,index[flowers],:], Flatness[index[flowers],:],Cum[:,index[flowers],:]=computeGridValue(tree,Count,lCount,IndexNear,index[flowers],mygrid, radius,Tloc[index[flowers]],weightingfunction)
                
                print('')
    return Mean_Std, Mom12, Flatness, Cum

#%% 
def MultiresQuantitySupportVD(data,radius, Nanalyse=2**14,destination=np.array([]), NonUniformData=False):
    """
     Compute multiscale quantities of the support of  a list of coordinate. 
         - box counting of the finite values  N(r) for each point.
         - Hat wavelet coefs (2* N(r) -N(np.sqrt(2)*r) of the set of finite values 
                                  
    
    Count, HatSupport = MultiresQuantitySupportVD(data,radius, Nanalyse=2**14,destination=np.array([]), NonUniformData=False)
    
    Parameters
    ----------
    data : numpy array of float
        Two dimensional data set of shape  (N, 2) containing the position x and y of N points.
    radius : numpy array of  float
        vector of radius values of shape (Nr,).
    Nanalyse : integer, optional
        Size of the batch. The default is 2**14.
    destination : TYPE, optional
        DESCRIPTION. The default is np.array([]).
    NonUniformData : boolean, optional
        Set to true if the data are not uniformly sampled is space. 
        If true, run faster using optimized batch contents.
        The default is False.

    Raises
    ------
    TypeError 'This is for point process ONLY.
        The data is certainly an image and this function is ONLY for non uniformly sampled data.

    Returns
    -------
    Count : numpy array of float
        Box counting of the number of points in a ball of radius radius around each data points.
        Count.shape = (N, Nr).
    HatSupport : numpy array of float
        Hat wavelet coefficients of the support : 2 * N(r) - N(sqrt(2) * r ).
        Count.shape = (N, Nr).

    """
    # check input data
    isimage, ismarked, Nmark = checkinputdata(data)
    if isimage:
        raise TypeError('This is for point process ONLY.')
    
    # we compute Box Counting B(r) nad B(r)-B(sqrt(2)*r)
    Count1 = localWaveTrans(data, radius, Nanalyse = Nanalyse, destination = destination, NonUniformData = NonUniformData)
    Count2 = localWaveTrans(data, radius*np.sqrt(2), Nanalyse = Nanalyse, destination = destination, NonUniformData = NonUniformData, verbose = False)
    
    Count = Count1
    HatSupport = 2 * Count1 - Count2
        
    return Count, HatSupport

#%% MultiresQuantityMarkVD
def MultiresQuantityMarkVD( data, radius, Nanalyse = 2**14, destination = np.array([]), NonUniformData=False):
    # check input data
    isimage, ismarked, Nmark = checkinputdata( data)
    if isimage:
        raise TypeError('This is for point process ONLY.')
    
    Poor1, Count1 = localWaveTrans(data, radius, Nanalyse = Nanalyse)
    Poor2, Count2 = localWaveTrans(data, radius * np.sqrt(2), Nanalyse = Nanalyse, verbose = False)
    
    print(' ---- Computing leader of coefs.')
    Count = Count1[ :, :, 0]
    HatSupport = 2 * Count1[ :, :, 0] - Count2[ :, :, 0]
    Mean = Count1[ :, :, 1]
    HatMean = Count1[ :, :, 1] - Count2[ :, :, 1]
    LHatMean = localLeadersP(data[:, (0, 1)], HatMean, radius, Nanalyse = 2**14, verbose = False)
    HatStd = Count1[:,:,2] - Count2[:,:,2]
    Poor = Poor1[ :, :, 1]   
    LPoor = localLeadersP( data[:, (0, 1)], Poor, radius, Nanalyse = 2**14, verbose = False)
    Std = Count1[ :, :, 2]
    
    
    return Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd     
        
#%%  MultiresQuantityVD
def MultiresQuantityVD(data,radius, Nanalyse=2**14,destination=np.array([]), NonUniformData=False, fftplan=[]):
    """
    

    Parameters
    ----------
    data : TYPE
        DESCRIPTION.
    radius : TYPE
        DESCRIPTION.
    Nanalyse : TYPE, optional
        DESCRIPTION. The default is 2**14.

    Returns
    -------
    Count : TYPE
        DESCRIPTION.
    HatSupport : TYPE
        DESCRIPTION.
    HatMean : TYPE
        DESCRIPTION.
    HatStd : TYPE
        DESCRIPTION.
    Poor : TYPE
        DESCRIPTION.
    LogPoor : TYPE
        DESCRIPTION.

    """
    # check input data
    isimage, ismarked, Nmark = checkinputdata(data)
   
    # choose method
    if (ismarked == 0) & (not isimage) :        
        
        #waveletKnnSupport
        Count, HatSupport = MultiresQuantitySupportVD(data,radius, Nanalyse=Nanalyse,destination=destination, NonUniformData=NonUniformData)
        
        Mean = np.array([])
        HatMean = np.array([])
        LHatMean = np.array([])
        Poor = np.array([])
        LPoor = np.array([])
        Std = np.array([])
        HatStd = np.array([])
        
    
    else: # marked points
        if isimage:
            #grrr
            #Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd = computeHatCoefsConv2d(data, radius)
            Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd = conv2D.MultiresQuantityConv(data, radius, fftplan=[])
        else:
            #hhhhh
            Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd  = MultiresQuantityMarkVD(data,radius, Nanalyse=Nanalyse,destination=destination, NonUniformData=NonUniformData)
            
            
                
    return Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd

def MultiresQuantityConv1d(data, radius, wavelet ='poor'):
    if wavelet == 'harr':
        Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd  = conv2D.computeHarrCoefsConv1d(data, radius)
    else:# Hat
        Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd = conv2D.computeCoefsConv1d(data, radius)
    # manque moyenne, log de la moyenne , std et std Hat
    #elif wavelet == 'harr':
    #elif wavelet == 'hat':
    
    return Count, HatSupport, Mean, HatMean, LHatMean, Poor, LPoor, Std, HatStd

# %%
def computeGridValuebivar(tree, Count1, Count2, IndexNear, index, mygrid, radius, Tloc, weightingfunction):
    """

    rhoss = computeGridValuebivarSS(tree,Count,IndexNear,index,mygrid, radius, T, k=0):

    For marked and non-marked point process
    Compute weighted statistics of the number of point location or value.
    The (geographical) weighting is obtained using the bi-squared function of size T.
       
    Input :

       tree      - a kdtree obtained on the IndexNear points (a subset of the Npts total points)
       Count     - matrix of size Nptsxlength(radius) with wavelet coef
       IndexNear - location of the points used for the kdtree
       index     - index of the grid points under investigations
       mygrid    - all the grid points (size (Ngridpts, 2) with X and Y spatial location
       radius   - list of scales to be investigated
       T        - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       k        - number of min neighbors for the adaptive kernel smoothing. If k=0 : fixed kernel.
        

     Output :

       rhoss - self similar correlation coefficient (of size length(data) X length(radius)

    """
    
    Tmax = np.nanmax(Tloc)
    neighbors_i_fixed, dist_ie_fixed = tree.query_radius(mygrid[index,:], r = Tmax, count_only=False, return_distance=True, sort_results=False)
    print('.',end='')
    sys.stdout.flush()

    # non weighted coef
    tmp_fixed1=[ Count1[IndexNear[neighbors_i_fixed[igrid]],:] for igrid in range (len(index))]
    tmp_fixed2=[ Count2[IndexNear[neighbors_i_fixed[igrid]],:] for igrid in range (len(index))]

    Wfinal=[  weightingfunction(dist_ie_fixed[igrid], Tloc[igrid]) for igrid in range (len(index))]
   
    m01=np.array([ np.nansum(tmp_fixed1[igrid] * Wfinal[igrid], axis=0) for igrid in range(len(index))])
    m10=np.array([ np.nansum(tmp_fixed2[igrid] * Wfinal[igrid], axis=0) for igrid in range(len(index))])   
    m02=np.array([ np.nansum(tmp_fixed1[igrid]**2 * Wfinal[igrid], axis=0) for igrid in range(len(index))])
    m20=np.array([ np.nansum(tmp_fixed2[igrid]**2 * Wfinal[igrid], axis=0) for igrid in range(len(index))])
    m11=np.array([ np.nansum(tmp_fixed1[igrid] * tmp_fixed2[igrid] * Wfinal[igrid], axis=0) for igrid in range(len(index))])   
    var1=m02-m01**2
    var2=m20-m10**2
    rho= (m11-m01*m10)/np.sqrt(var1*var2)
    
    return rho
# 

# %% 
def LocalCorrAnalysisVD(data, Count1, Count2, X, Y, radius, T, adaptive=False, weights = 'flat', Nanalyse=2**16, Tloc=[]):
    """

   res = WaveSmoothingBivar(data,Wave,X,Y,radius,T,Nanalyse=2**16, k = 0)

   For marked and non-marked point process
   Compute kernel smoothing of the wavelet coefficient of a dataset of points.
   The geographical weighting is obtained using the bi-squared function of size T.
   Only for non valued analysis (non marked process)
   
   Input :

       data     - matrix of size Nx2 --> position (x,y) for N points
       Wave     - matrix of size Nxlength(radius) with wavelet count
                  Can be obtained using the function  GWFA_count.m
       X        - array of dim 2 with x-postion of the grid nodes
       Y        - array of dim 2 with y-postion of the grid nodes : X and Y must have the same size
       radius   - list of scales to be investigated
       T        - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       Nanalyse - number of points to analyse in one bach. Default is 2**16 points.
       k        - number of min neighbors for the adaptive kernel smoothing. If k=0 : fixed kernel.
       

   Output :

       res - list of two dimensional numpy matrix (of size length(data) X length(radius)

              res[Ø] = Wmean   : weighted mean
              res[1] = Wstd    : weighted standart deviation
              res[2] = NWratio : non weighted mean
              res[3] = NWstd   : non weighted standart deviation
              res[4] = Mom     : weighted moment (order 0 to 4) of the absolute value
                      of thecoefficient. matrix of size 5 X length(data) X length(radius)
              res[5] = Cum     : weighted cumulant one and two 
                      of thecoefficient. matrix of size 2 X length(data) X length(radius)

    """
    assert(Count1.shape == Count2.shape)
    
    si=data.shape
    if si[1] < 2 or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')

    if Count1.ndim > 2:
        raise TypeError('The second arument must be two or threedimensional.')
   
    if Count1.shape[0] != si[0]:
        raise TypeError('The two first arument must have the same length.')

    if radius[radius>T].shape[0] >0:
        raise TypeError('The last argument must be greater than a sqrt(2)*radius.')
    # check grid
    if np.sum(np.abs(np.array(X.shape)-np.array(Y.shape)))>0:
        raise TypeError('X and Y must have the same size.')

    if np.sum(np.diff(X,axis=1,n=2))>0:
        raise TypeError('X must be regulary sampled.')
    else:
        gridsizex=max(X[0][1] - X[0][0],X[1][0] - X[0][0])
        
    if np.sum(np.diff(Y,axis=0,n=2))>0:
        raise TypeError('Y must be regulary sampled.')
    else:
        gridsizey=max(Y[1][0] - Y[0][0],Y[0][1] - Y[0][0])
    
    if gridsizex != gridsizey:
        raise TypeError('X and Y must have same sampling.')
    
    if radius.size != Count1.shape[1]:
        #print(radius.size, Count.shape[1])
        raise TypeError('The size of the second arument must be [N,R,M] where R is the number of scales (=radius.size)')
 
    # create grid points
    mygrid=np.column_stack([X.flatten(),Y.flatten()])
    gridlength=mygrid.shape[0]
    gridsizecurr = gridsizex


    Npairs=1
    print('There is {:d} different pair(s)'.format(Npairs))
    
    Nanalyse2=300000
    if Nanalyse==0:
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=gridlength
        Nbunch=1
    else: # find how many bunchs        
        Nbunch = int(np.ceil(gridlength/Nanalyse))
    
    # choose weighting  
    if isinstance(weights, str):
        if weights == 'flat':
            weightingfunction = partial(geographicalWeight,func = flatWindow)
            print('No weight')
        elif weights == 'Epanechnikov':
            weightingfunction = partial(geographicalWeight,func = EpanechnikovWindow)
        elif weights == 'bisquare':
            weightingfunction = partial(geographicalWeight,func = bisquareWindow)
        elif weights == 'triangular':
            weightingfunction = partial(geographicalWeight,func = triangularWindow)
        elif weights == 'tricube':
            weightingfunction = partial(geographicalWeight,func = tricubeWindow)            
    elif isinstance(weights, types.FunctionType):
        weightingfunction = partial(geographicalWeight,func = weights)
    else:
        raise TypeError('ERROR : weighting is of unknown type.).')

    if adaptive == True:
        if len(Tloc) ==0: # compute local env
            print('ADAPTIVE KERNEL: Computing pilot density')
            CountT= localWaveTrans(data[:,(0,1)], np.atleast_1d(T))
            WmeanStd_T, Mom_T, Flat_T, Cum_T = LocalMsAnalysisVD(data[:,(0,1)],CountT,X,Y,np.atleast_1d(T), T=T,Nanalyse=2**9)
            Wmean_Tloc = WmeanStd_T[0,:,0]
            G = np.nanmean(Wmean_Tloc)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                lambda_ie = G/Wmean_Tloc
            Tloc = lambda_ie*T
            Tloc[np.isinf(Tloc)] = T
            Tlocmax = np.nanquantile(Tloc,0.95)#.9
            Tloc[Tloc>Tlocmax] = Tlocmax 
            Tlocmin = np.nanquantile(Tloc,0.05)#.05
            Tloc[Tloc<Tlocmin] = Tlocmin 
            T=Tlocmax
            print('ADAPTIVE KERNEL computed.')
        elif X.flatten().shape[0] == Tloc.flatten().shape[0]:
            T=np.nanmax(Tloc)
        else: 
            raise TypeError('ERROR : wrong size for the local environment')
    else:
        Tloc=T*np.ones(X.size)

    # results allocations
    rho = np.nan*np.zeros( (gridlength,radius.size), dtype = float)
       
    # loop on the bunch
    for ibunch in range(Nbunch):
        # %
        print('bunch {:d}/{:d} '.format(ibunch+1,Nbunch), end=' ')
        sys.stdout.flush()
        # get data index of current bunch
        index=np.arange(ibunch*Nanalyse,(ibunch+1)*Nanalyse,1)
        index=index[index<gridlength]

        # we restrict the tree to points whithin a radius T 
        mi=np.min(mygrid[index,:], axis=0)
        ma=np.max(mygrid[index,:], axis=0)
        IndexNear=np.where((data[:,0] >mi[0]-T) & (data[:,0] <ma[0]+T) & (data[:,1]  >mi[1]-T) & (data[:,1] <ma[1]+T))[0]

        # make the tree with the nearest points only
        if IndexNear.shape[0]>0:
            tree = KDTree(data[IndexNear,0:2])
            Idxtot = tree.query_radius(mygrid[index,:], r = np.sqrt(2)*gridsizecurr, count_only=False, return_distance=False)
        
        
            # find and remove empty (no point location) grid pixels
            thelengths=np.array([ Idxtot[igrid].shape[0] for igrid in range (len(Idxtot))])
            IdxMin, = np.where(thelengths>0.)
            index = index[IdxMin]
            Idxtot=Idxtot[IdxMin]
            thelengths=thelengths[IdxMin]
        
            # Managed the number of grid points  (too many neighboors for some)
            cumsumbunch=np.cumsum(thelengths)
            Nflowers=int(np.ceil(np.sum(thelengths)/Nanalyse2))
            print('{:d} batch(s).'.format(Nflowers))
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                for iflower in range(Nflowers):
                    i1,=np.where(iflower*Nanalyse2 <= cumsumbunch)
                    i2,=np.where(cumsumbunch <  (iflower+1)*Nanalyse2)
                    flowers=np.intersect1d(i1,i2)
                    rho[index[flowers],:] = computeGridValuebivar(tree,Count1,Count2,IndexNear,index[flowers],mygrid, radius,Tloc[index[flowers]],weightingfunction)
        print('.')   
       
    return rho

#%% 
def AllLocalQuantities(data, coefs, radius, Nanalyse = 2**14,  NonUniformData = False, verbose = True):
    """
    
     [Count, Wave, CountG] = aveTrans(source, radius, T=None, Nanalyse=2**16, destination = []))
    
    
    
    Compute box-counting and wavelet coefficient on a valued/non valued set of data points.
    
    If the data are not valued, count for every data point
            -- the number of neighboors in ball of radius r (GWFA) : N(r).
            -- the wavelet coeficient at scale r (GWMFA) : 2*N(r)-N(sqrt(2)*r).
    
    If the data are valued, count for every datapoint
            -- the number of neighboors in ball of radius r, the mean and std of the marked value
            -- the wavelet coeficient at scale r on the marked value.
    
    Input :
    
        source     - Non-marked  point process : matrix of size N X 2 for N points
                        where source[i,:]=[X_i,Y_i] with 2D cooprdonate of point i.
                      Marked  point process :  matrix of size Nx3
                        where source[i,:]=[X_i,Y_i, mark_i] with 2D cooprdonate of point i and value
        radius      - list of scales to be investigated
        Nanalyse    - number of points to analyse in one bach. Default is 2**16 points.
                        If Nanalyse=0, compute all the points in once (dangerous!!!)
        destination - Non marked point process (destination_i=[X_I,Y_i]) where the coeficient are calculated
                        Default empty : compute at source position.
    Output :
    
        Count    - matrix of size Nxlength(radius) with box-counting coefficients
        Wave     - matrix of size Nxlength(radius) with wavelet coefficients
        Count    - matrix of size Nx2 with box-counting and wavelet coeficient at scale T
                  
    
    
    Usage exemple :
    

    """
 
    #% check input data
    si = data.shape
    if si[1] < 2: # or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
         
    destination = np.copy(data[:,0:2])
    #val = data[:,2]
    #valtot = data[:,2:]
    #data = data[:,0:2]
            
    radius = np.atleast_1d(radius)
    radiustot = np.sort(radius)
    #temp,temp2,indexreadius=np.intersect1d(radius,radiustot, return_indices=True)

    # some constants
    scalemax=np.max(radiustot)
    N = destination.shape[0]
    
    # sub parameters to partition  the data
    if Nanalyse==0: 
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=N
        Nbunch=1
    else:
        # find how many bunchs
        if NonUniformData:
            centers, sizeblock = getoptimblock(destination, Nanalyse)
            Nbunch = len(sizeblock)
            if verbose:
                print('Block optimisation for non uniformly spaced data.')
                       
        else:
            sizetmp = Nanalyse
            temp= destination // sizetmp
            block, count = np.unique(temp, axis=0, return_counts=True)
            centers = block*sizetmp+sizetmp//2
            sizeblock = count*0+sizetmp
            Nbunch = len(block)
            if verbose:
                print('No block optimisation uniformly spaced data.')
            
    
    Nanalyse2 = 2**22
    if verbose:
        print('Computation in {:d} bunches :'.format(Nbunch),end=' ')  
    
   
    # allocation
    Count = np.nan*np.zeros((N,radius.shape[0],7),dtype=float)
    
    
    # set the workers for knn search
    neigh = NearestNeighbors(n_jobs=4)
    
    #% loop on the bunches
    for ibunch in range(Nbunch):
        #%
        print(ibunch+1, end=' ')
        sys.stdout.flush()
        #
        center = centers[ibunch,:]
        sizetmp = sizeblock[ibunch]
        #
        index, = np.where(((destination[:,0] >= center[0]-sizetmp//2) & (destination[:,0]< center[0]+sizetmp//2)) & ((destination[:,1]>= center[1]-sizetmp//2) & (destination[:,1]< center[1]+sizetmp//2)))
        IndexNear, = np.where(((destination[:,0] >= center[0]-(sizetmp//2+scalemax+1)) & (destination[:,0]< center[0]+(sizetmp//2+scalemax+1))) & ((destination[:,1]>= center[1]-(sizetmp//2+scalemax+1)) & (destination[:,1]< center[1]+(sizetmp//2+scalemax+1))))
        neigh.fit(destination[IndexNear,:])
               
        Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2], radius = scalemax, return_distance = True)
        Maxlength = max(map( len, IdxTot))        
        Nblock = int(np.ceil(index.shape[0] * Maxlength / Nanalyse2))           
        Nptx = len(index) // Nblock + 1
        if (Nblock - 1)*Nptx > min(len(index), Nblock * Nptx):
            Nblock = Nblock - 1
        
        malist2=[(IdxTot[i*Nptx:min(len(index),(i+1)*Nptx)], Disttot[i*Nptx:min(len(index),(i+1)*Nptx)]) for i in range(Nblock)]
        partialknncountbis = partial(threaded_All_Valued, radiustot = radiustot,val = coefs[IndexNear,:])      
        #%
        with ThreadPoolExecutor() as executor:
            result_list3 = executor.map(partialknncountbis, malist2)
        
        Count[index,:,:] = np.vstack(list(result_list3))
            
            
    del neigh
    print('.')
          
     
    return Count
    
#%%
def AllLocalQCorrCoef(data, coefs1, coefs2, radius, Nanalyse = 2**14,  NonUniformData = False, verbose = True):
    """
    
     [Count, Wave, CountG] = aveTrans(source, radius, T=None, Nanalyse=2**16, destination = []))
    
    
    
    Compute box-counting and wavelet coefficient on a valued/non valued set of data points.
    
    If the data are not valued, count for every data point
            -- the number of neighboors in ball of radius r (GWFA) : N(r).
            -- the wavelet coeficient at scale r (GWMFA) : 2*N(r)-N(sqrt(2)*r).
    
    If the data are valued, count for every datapoint
            -- the number of neighboors in ball of radius r, the mean and std of the marked value
            -- the wavelet coeficient at scale r on the marked value.
    
    Input :
    
        source     - Non-marked  point process : matrix of size N X 2 for N points
                        where source[i,:]=[X_i,Y_i] with 2D cooprdonate of point i.
                      Marked  point process :  matrix of size Nx3
                        where source[i,:]=[X_i,Y_i, mark_i] with 2D cooprdonate of point i and value
        radius      - list of scales to be investigated
        Nanalyse    - number of points to analyse in one bach. Default is 2**16 points.
                        If Nanalyse=0, compute all the points in once (dangerous!!!)
        destination - Non marked point process (destination_i=[X_I,Y_i]) where the coeficient are calculated
                        Default empty : compute at source position.
    Output :
    
        Count    - matrix of size Nxlength(radius) with box-counting coefficients
        Wave     - matrix of size Nxlength(radius) with wavelet coefficients
        Count    - matrix of size Nx2 with box-counting and wavelet coeficient at scale T
                  
        
    """
 
    #% check input data
    si = data.shape
    if si[1] < 2: # or si[1]>3:
        raise TypeError('The second dimension of first argument must be of length 2 or 3.')
         
    destination = np.copy(data[:,0:2])           
    radius = np.atleast_1d(radius)
    radiustot = np.sort(radius)

    # some constants
    scalemax=np.max(radiustot)
    N = destination.shape[0]
    
    # sub parameters to partition  the data
    if Nanalyse==0: # this is dangerous
        print('Warning : dangerous if your set of data is large.')
        Nanalyse=N
        Nbunch=1
    else:
        # find how many bunchs
        if NonUniformData:
            centers, sizeblock = getoptimblock(destination, Nanalyse)
            Nbunch = len(sizeblock)
            if verbose:
                print('Block optimisation for non uniformly spaced data.')
                       
        else:
            sizetmp = Nanalyse
            temp= destination // sizetmp
            block, count = np.unique(temp, axis=0, return_counts=True)
            centers = block*sizetmp+sizetmp//2
            sizeblock = count*0+sizetmp
            Nbunch = len(block)
            if verbose:
                print('No block optimisation uniformly spaced data.')
            
    
    Nanalyse2 = 2**22
    if verbose:
        print('Computation in {:d} bunches :'.format(Nbunch),end=' ')  
    
   
    # allocation
    Count = np.nan*np.zeros((N,radius.shape[0],2),dtype=float)
    
    
    # set the worker for knn search
    neigh = NearestNeighbors(n_jobs=4)
    
    #% loop on the bunches
    for ibunch in range(Nbunch):
        #%
        print(ibunch+1, end=' ')
        sys.stdout.flush()
        #
        center = centers[ibunch,:]
        sizetmp = sizeblock[ibunch]
        #
        index, = np.where(((destination[:,0] >= center[0]-sizetmp//2) & (destination[:,0]< center[0]+sizetmp//2)) & ((destination[:,1]>= center[1]-sizetmp//2) & (destination[:,1]< center[1]+sizetmp//2)))
        IndexNear, = np.where(((destination[:,0] >= center[0]-(sizetmp//2+scalemax+1)) & (destination[:,0]< center[0]+(sizetmp//2+scalemax+1))) & ((destination[:,1]>= center[1]-(sizetmp//2+scalemax+1)) & (destination[:,1]< center[1]+(sizetmp//2+scalemax+1))))
        neigh.fit(destination[IndexNear,:])
               

        Disttot, IdxTot = neigh.radius_neighbors(destination[index,0:2], radius = scalemax, return_distance = True)
        Maxlength = max(map( len, IdxTot))
        
        Nblock = int(np.ceil(index.shape[0] * Maxlength / Nanalyse2))           
        Nptx = len(index) // Nblock + 1
        if (Nblock - 1)*Nptx > min(len(index), Nblock * Nptx):
            Nblock = Nblock - 1        
        malist2=[(IdxTot[i*Nptx:min(len(index),(i+1)*Nptx)], Disttot[i*Nptx:min(len(index),(i+1)*Nptx)]) for i in range(Nblock)]
        partialknncountbis = partial(threaded_corr_coef, radiustot = radiustot,val1 = coefs1[IndexNear,:], val2 = coefs2[IndexNear,:])
        
        #%
        with ThreadPoolExecutor() as executor:
            result_list3 = executor.map(partialknncountbis, malist2)
        
        Count[index,:,:] = np.vstack(list(result_list3))
            
            
    del neigh
    print('.')
          
     
    return Count
    

    
    


"""
======================================================
2D Convolution package for LomPy (:mod:`lompy.conv2D`)
======================================================
          
.. currentmodule: lompy.conv2D

The package features yield the multiresolution quantities and the local 
scaling parameters  for uniformly distributed (raster) 2D data


Overview
--------
Routines for the muliresolution quantities:
    
    lompy.conv2D.MultiresQuantityRD
    
Routines for the univariate local analysis:
    
    lompy.conv2D.LocalMsAnalysisRD

Routines for the bivariate local analysis:
    
    lompy.conv2D.LocalCorrAnalysisRD

"""

#Copyright (C) 2023 GNU AGPLv3, LomPy 2023, S.G. Roux and J. Lengyel.
#All rights reserved.
#Contact: stephane.roux@ens-lyon.fr, jankalengyel@gmail.com
#Other Contributors: P. Thiraux


import warnings
import numpy as np
import pyfftw
import multiprocessing

#%% 
pyfftw.interfaces.cache.enable()
pyfftw.config.NUM_THREADS = multiprocessing.cpu_count()
N_threads = multiprocessing.cpu_count()

#%%
def prepare_plan2d(Nx, Ny, N_threads = N_threads):
    """
    create plan for fftw
    
    a,fft_a, fft_object, b,ffti_b, ffti_object = prepare_plan2d(Nx, Ny, N_threads = N_threads)

    Parameters
    ----------
    Nx : integer
        x dimension of the plan.
    Ny : integer
        y dimension of the plan
    N_threads : integer, optional
        Number of thread to use. The default is multiprocessing.cpu_count.

    Returns
    -------
    a : numpy array of float64
        aligned memory.
    fft_a : numpy array of complex128
        aligned memory for fft.
    fft_object : fftw object.
        FORWARD fft_object.
    b : numpy array of float64
        aligned memory.
    ffti_b : numpy array of complex128
        aligned memory for fft.
    ffti_object : fftw object.
        BACKWARD' fft_object.

    """
    a     = pyfftw.empty_aligned((Nx, Ny), dtype='float64')
    fft_a = pyfftw.empty_aligned((Nx, Ny//2+1), dtype='complex128')
    fft_object = pyfftw.FFTW(a, fft_a, axes=(0,1), direction='FFTW_FORWARD', threads=N_threads)
        
    ffti_object = pyfftw.FFTW(fft_a, a, axes=(0,1), direction='FFTW_BACKWARD', threads=N_threads)
    return  fft_object, ffti_object

#%%
def do_fftplan2d( N, scales, N_threads = N_threads, wavelet = 'poor'):
    """
    Create fftw plan for fast wavelet transform
    
    a,fft_a, fft_object, b,ffti_b, ffti_object = do_fftplan2d( N, scales, N_threads = N_threads, wavelet = 'poor')

    Parameters
    ----------
    N : integer
        DESCRIPTION.
    scales : float
        vector of positive float.
    N_threads : integer, optional
        Number of thread to use. The default is multiprocessing.cpu_count..
    wavelet : TYPE, optional
        Wavelet to use. The default is 'poor'.

    Returns
    -------
    a : numpy array of float64
        aligned memory.
    fft_a : numpy array of complex128
        aligned memory for fft.
    fft_object : fftw object.
        FORWARD fft_object.
    b : numpy array of float64
        aligned memory.
    ffti_b : numpy array of complex128
        aligned memory for fft.
    ffti_object : fftw object.
        BACKWARD fft_object.

    """
    
    N1, N2 = N
    scales = np.atleast_1d(scales)
    if len(scales) == 1:
        maxradius = int(np.ceil(np.max(scales)))
    else:
        if np.char.equal(wavelet,'poor'):
            maxradius = int(np.ceil(np.max(scales)))
        else:
            maxradius = int(np.ceil(np.sqrt(2)*np.max(scales)))
    
    # add border
    N1 = N1 + 2 * maxradius
    N2 = N2 + 2 * maxradius
    Nk = 2 * maxradius + 1
    
    Nfft1 = N1 + Nk - 1
    Nfft2 = N2 + Nk - 1
    fft_object, ffti_object = prepare_plan2d(Nfft1, Nfft2, N_threads)

    return fft_object, ffti_object
#%
#%% 
def convFFTW(a, TFW, fft_object,ffti_object):
    """
    compute the inverse Fourier transform of the product of  
    FFT(a), the Fourier tranform of a,  by TFW. 
    
    temp2 = convFFTW(a, TFW, fft_object,ffti_object)
    
    Parameters
    ----------
    a : numpy array of float64
        signal or image.
    TFW : numpy array of complex128
        Fourier transform of a kernel.
    fft_object : fftw object.
        fftw object to use for FORWARD direction.
    ffti_object : ftw object.
        fftw object to use for BACKWARD direction.

    Returns
    -------
    temp2 : float64
        results.

    """
    b = fft_object(a)
    
    ar=ffti_object(b*TFW)
    temp2=np.copy(ar)
    return temp2
#%% 
def checkinputdata(data):
    """
    Check the type of data.
    Image (uniformly sampled)?
    Marked?
    How many marks?
    
    isimage, ismarked, Nmark = checkinputdata(data)

    Parameters
    ----------
    data : numpy array of float
        The data to check.

    Returns
    -------
    isimage : Boolean
        Equal to 1 if the data is an image.
    ismarked : Boolean
        Equal to 1 if the data ia a marked point process.
    Nmark : int
        Number oif mark in the data.

    """
    # check input data
    si=data.shape
    if len(si)==1:
        print('Error : the input argument must have at least two dimensions')
        return
    else:
        
        if (si[0]>100) & (si[1]>100):
            isimage=1
            if len(si)>2:
                #print('The input seems to be a set of  images  of size ({:d},{:d}))'.format(si[2]),si[0])
                print('The input seems to be a set of  {:d}  images  of size ({:d},{:d})).'.format(si[2],si[0],si[1]))
                ismarked = 1
                Nmark = si[2]
            else:
                print('The input seems to be an image of size ({:d},{:})'.format(si[0],si[1]))
                ismarked = 0
                Nmark = 0
        else:
            isimage=0
            if si[1]>2:
                print('The input seems to be a marked point process with {:} points and {:d} marks.'.format(si[0],si[1]-2))
                
                ismarked = 1
                Nmark = si[1]-2
            else:
                print('The input seems to be a marked point process with {:} points.'.format(si[0]))
                ismarked = 0
                Nmark = 0
    return isimage, ismarked, Nmark

#%%
def threshkernelangle(kernel,thetaTot,theta,dtheta):
    
    if theta+dtheta/2 > np.pi:
        kernel[(thetaTot > theta+dtheta/2-2*np.pi) & (thetaTot < theta-dtheta/2)]=0
    elif theta-dtheta/2 < -np.pi:
        kernel[(thetaTot < theta-dtheta/2+2*np.pi) & (thetaTot > theta+dtheta/2)]=0
    else:
        kernel[thetaTot <= theta-dtheta/2]=0
        kernel[thetaTot > theta+dtheta/2]=0    
    return kernel
# %%  geo weighting
def EpanechnikovWindow(z):
    return 0.75*(1-z**2) 

def triangularWindow(z):
    return (1 - np.abs(z)) 

def tricubeWindow(z):
    return (70/81)*(1-np.abs(z)**3)**3

def bisquareWindow(z):
    return (1-z**2)**2 

def flatWindow(z):
    return np.ones(z.shape)
#%
def geographicalWeight(dd,T,func):
    """
    
    W = geographicalWeight(dist,T,Nr)
    
    return the weight according to the distance dist 
    using a local environment of T.
    
    Input :

       dist - distance of the point. dist and index as the shame shape
       T    - bandwith of fixed kernel OR distance upper boundary of adaptive kernel
       Nr   - number of replication of the weight
        
     Output :

       W - the weight. Array of size (len(dd),Nr)
   
    
    ##
     S.G.  Roux, ENS Lyon, December 2020,  stephane.roux@ens-lyon.fr
     J.L June 2021
     S.G.  Roux, ENS Lyon, November 2021
    """  

    z = dd/T 
    W=func(z)
    W[dd>=T]=0
    W=W/np.sum(W,0)
    W=W[:,np.newaxis]
    return W

#%% 
def MultiresQuantityRD(image, radiustot, fftplan=[], theta=[], dtheta=[], p=2):
    """
    Compute multiscale quantities from Image. The image can contain non finite values.
        - box counting of the finite values  N(r) for each pixel
        - Hat wavelet coefs (2* N(r) -N(np.sqrt(2)*r) of the set of finite values 
        - local mean, std, Hat wavelet coeficient and Poor wavelet coefficient 
        of the image (discarding non finite values) 
    
    Count, HatSupport, Mean, LMean, HatMean, LHatMean, Poor, LPoor, Std, HatStd = MultiresQuantityRD(image, radiustot, fftplan=[])


    Parameters
    ----------
    data : numpy array of float
        Two dimensional data set of shape  (Nx, Ny).
    radius : numpy array of  float
        vector of radius values of shape (Nr,).
    fftplan : fftw object, optional
        For fast computation we can prepare fftw plan using function
        do_fftplan2d. The default is [].

    Raises
    ------
    TypeError This is for image ONLY.
        The input data is not an image : a numpy array of size (Nx,Ny).

    Returns
    -------
    Count : numpy array of float
        box counting coeficient of the support: numpy array of  shape  (Nr, Nx, Ny).
        It's the number of non nan values in a ball of radius r : N_x(r)
    HatSupport : numpy array of float
        Hat wavelet coeficient of the support: numpy array of  shape  (Nr, Nx, Ny).
        Defined as  2*N_x(r)-N_x(np.sqrt(2)*r)
    Mean : numpy array of float
        Local average (on a ball of size radius) of the mark : numpy array of  shape  (Nr, Nx, Ny).
        M_x(r)
    LMean : numpy array of float
        Local average of the logarithm of the mark : numpy array of  shape  (Nr, Nx, Ny).
    HatMean : numpy array of float
        Hat wavelet coeficient of the mark : numpy array of  shape  (Nr, Nx, Ny).
        Defined as  M_x(r)-M_x(np.sqrt(2)*r)
    LHatMean : numpy array of float
        local Logarithm of Hat wavelet coeficient of the mark : numpy array of  shape  (Nr, Nx, Ny).
    Poor : numpy array of float
        Poor wavelet coeficient of the mark : numpy array of  shape  (Nr, Nx, Ny).
        Defined as  data_x-M_x(r)
    LPoor : numpy array of float
        local Logarithm of  Poor wavelet coeficient of the mark : numpy array of  shape  (Nr, Nx, Ny).      
    Std : numpy array of float
        Local standart deviation (on a ball of size radius)  of the mark.
        numpy array of  shape  (Nr, Nx, Ny). 
    HatStd : numpy array of float
        Local standart deviation (on a ball of size radius)  of the Hat wavelet coeficients. 
        numpy array of  shape  (Nr, Nx, Ny). 

    """
    isimage, ismarked, Nmark = checkinputdata(image)
    if not isimage:
        raise TypeError('This is for image ONLY.')
    
    isAni = False
    if len(dtheta)+len(theta) == 2: # anisotropic case
        isAni = True
    image[np.isinf(image)] = np.nan
    maxradius = int(np.ceil(np.sqrt(2)*np.max(radiustot)))
    im = np.pad(image, ((maxradius, maxradius) ,(maxradius, maxradius)), constant_values=((np.nan, np.nan),(np.nan, np.nan)))
    
    IndNan = np.isnan(im)
    imcopy = np.copy(im)
    imcopy[IndNan] = 0
    im01 = np.ones(im.shape)
    im01[IndNan] = 0
    
    N1,N2 = imcopy.shape
    Nborder = int(np.ceil(np.sqrt(2)*np.max(radiustot)))
    Nk = 2*Nborder+1
    
    # define distance for filter
    y, x = np.ogrid[-Nborder:Nborder+1, -Nborder:Nborder+1]
    d = x**2 + y**2
    thetaTot = np.flip(np.arctan2(y,x),axis=0)
    
    # fftw plan
    Nfft1 = N1+Nk-1
    Nfft2 = N2+Nk-1
    
    if len(fftplan)==0:
        fft_object,  ffti_object = prepare_plan2d(Nfft1, Nfft2, N_threads)
    else:
        fft_object,  ffti_object = fftplan
   

    # allocations
    HatSupport = np.zeros((radiustot.shape[0], N1, N2))
    HatStd = np.zeros((radiustot.shape[0], N1, N2))
    Mean = np.zeros((radiustot.shape[0], N1, N2))
    HatMean = np.zeros((radiustot.shape[0], N1, N2))
    Poor  = np.zeros((radiustot.shape[0], N1, N2))
    Std = np.zeros((radiustot.shape[0], N1, N2))
    Count = np.zeros((radiustot.shape[0], N1, N2))
    
    LeadPoor = np.zeros((radiustot.shape[0], N1, N2))
    LeadHat = np.zeros((radiustot.shape[0], N1, N2))
    # loop on scales
    for ir in range(radiustot.shape[0]):
        # %
        print(ir, end=' ', flush = True)
        # first scale
        kernel = np.zeros((Nk,Nk))
        kernel[np.sqrt(d) <= radiustot[ir]]=1
        if isAni:
            kernel=threshkernelangle(kernel,thetaTot,theta,dtheta)
        TFW = np.copy(fft_object(np.pad(kernel, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
        # count finite coefs
        imtmp = np.pad(im01, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid1 = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]       
        CountValid1[CountValid1 < 0.1] = np.nan # must be integer >0
        Cc = np.copy(CountValid1)
        Cc[IndNan] = np.nan
        # mean
        imtmp = np.pad(imcopy, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]
        lamean = lasum / CountValid1
        lamean[IndNan] = np.nan
        # moment of order 2 and std
        lasum = convFFTW(imtmp**2, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]      
        M2 = lasum / CountValid1
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning)
            lavar = np.maximum(M2-lamean**2,0)
            lastd = np.sqrt(lavar)   
        lastd[IndNan] = np.nan
        
        # Allocations
        Count[ir,:,:] = Cc
        Poor[ir,:,:] = lamean - im
        Mean[ir,:,:] = lamean
        Std[ir,:,:] = lastd

        
        # 2nd scales : for hat wavelet
        kernel = np.zeros((Nk,Nk))
        kernel[(np.sqrt(d) <= np.sqrt(2)*radiustot[ir]) & (np.sqrt(d) > radiustot[ir])] = 1
        if isAni:
            kernel=threshkernelangle(kernel,thetaTot,theta,dtheta)
        TFW = np.copy(fft_object(np.pad(kernel, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
        # count finite coefs
        imtmp = np.pad(im01, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid2 = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]       
        CountValid2[CountValid2 < 0.1] = np.nan 
        Cc2 = np.copy(CountValid2)
        Cc2[IndNan] = np.nan
        # mean
        imtmp = np.pad(imcopy, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum2 = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]
        lamean2 = lasum2 / CountValid2
        lamean2[IndNan] = np.nan
        # moment of order2
        lasum = convFFTW(imtmp**2, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]
        M2 = lasum / CountValid2
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning)
            lavar = np.maximum(M2 - lamean**2, 0)
            lastd2 = np.sqrt(lavar)              
        lastd2[IndNan] = np.nan
        # allocation
        HatSupport[ir,:,:] = Cc - Cc2
        HatMean[ir,:,:] = lamean - lamean2       
        HatStd[ir,:,:] = lastd - lastd2
        
        # Poor  Leaders
        Leadtmp=np.abs(Poor[ir,:,:])
        isnanLead = np.isnan(Leadtmp)
        
        kernel = np.zeros((Nk,Nk))
        kernel[np.sqrt(d) <= radiustot[ir]]=1
        TFW = np.copy(fft_object(np.pad(kernel, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
        
        tmp = np.ones(Leadtmp.shape)
        tmp[isnanLead] = 0
        imtmp = np.pad(tmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]       
        CountValid[CountValid < 0.1] = np.nan
        Leadtmp[isnanLead] = 0
        imtmp = np.pad(Leadtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum = convFFTW(imtmp**p, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning)
            lasum = lasum**(1/p) #/ CountValid
            #### ITT
            lasum[isnanLead]= np.nan 
        LeadPoor[ir,:,:] = lasum 
        
        # Poor  Hat
        Leadtmp=np.abs(HatMean[ir,:,:])
        isnanLead = np.isnan(Leadtmp)
        tmp = np.ones(Leadtmp.shape)
        tmp[isnanLead] = 0
        imtmp = np.pad(tmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]       
        CountValid[CountValid < 0.1] = np.nan
        Leadtmp[isnanLead] = 0
        imtmp = np.pad(Leadtmp**p, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum = convFFTW(imtmp, TFW, fft_object,ffti_object)[Nborder:Nfft1-Nborder, Nborder:Nfft2-Nborder]
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = RuntimeWarning)
            lasum = lasum**(1/p)
            ####ITT
            lasum[isnanLead]= np.nan 
        LeadHat [ir,:,:] = lasum 

        
        
    print('.')         
    HatMean = HatMean[:, maxradius:-maxradius, maxradius:-maxradius]
    Poor = Poor[:, maxradius:-maxradius, maxradius:-maxradius]
    Mean = Mean[:, maxradius:-maxradius, maxradius:-maxradius]
    Std = Std[:, maxradius:-maxradius, maxradius:-maxradius]    
    Count = Count[:, maxradius:-maxradius, maxradius:-maxradius]    
    HatStd = HatStd[:, maxradius:-maxradius, maxradius:-maxradius]    
    HatSupport = HatSupport[:, maxradius:-maxradius, maxradius:-maxradius]     
    LeadPoor = LeadPoor[:, maxradius:-maxradius, maxradius:-maxradius] 
    LeadHat = LeadHat[:, maxradius:-maxradius, maxradius:-maxradius]     
    Count = Count.transpose(1, 2, 0)
    HatSupport = HatSupport.transpose(1, 2, 0)
    Mean = Mean.transpose(1, 2, 0)
    HatMean = HatMean.transpose(1, 2, 0)
    Poor = Poor.transpose(1, 2, 0)
    Std = Std.transpose(1, 2, 0)
    HatStd = HatStd.transpose(1, 2, 0)
    LeadPoor = LeadPoor.transpose(1, 2, 0)
    LeadHat = LeadHat.transpose(1, 2, 0)
    return Count, HatSupport, Mean, HatMean, LeadHat, Poor, LeadPoor, Std, HatStd 
#%%
def bisquarekernel(x):
    '''
        return bisaquare kernel weigths (1-x**2)**2 ewith -1< x< 1
        
        
        w=isquarekernel(x)
        
        Input :
            x vector of value taken between -1 and 1.
            
        Output :            
        
            w weigths values
        
     ##

    '''

    A = (1-x**2)**2
    A[np.abs(x) > 1] = 0
    #ITT norm
    return A
#%%
def LocalMsAnalysisRD(coefs,lcoefs,L, window='flatcircle',fftplan=[], gridsize=1):
    '''
        Compute localized statistics (moments and cumulants).
        
        Mom1, Mom2, Flatness ,Cum1, Cum2 = LocalMsAnalysisRD(coefs,lcoefs,L,window)
        
        Inputs :
            
            coefs  : set of ceoficients used to compute moments and flatness
            
            lcoefs:  set of log coefficients used to compute the cumulants
            
            L : size of the average (integer stricty positif)
            
            window : weighting function to use. Default : window='flat'.
                     can be window='bisquare'
            
        Outputs :   
            
            Mom1, Mom2 : localized moment of order one and two
            
            Flatness : localized flatness
            
            cum1, cum2 : localized log cumulant of order one and two
            
        
     ##

    '''
    
    assert(coefs.shape == lcoefs.shape)
    # transpose axis
    coefs = coefs.transpose(2, 0, 1)
    lcoefs = lcoefs.transpose(2, 0, 1)
    
    # take care of border effect --> pad by nan nan 
    imCoefs = np.pad(coefs, ((0, 0), (L, L) ,(L, L)), constant_values=((np.nan, np.nan), (np.nan, np.nan), (np.nan, np.nan)))
    imLCoefs = np.pad(lcoefs, ((0, 0), (L, L) ,(L, L)), constant_values=((np.nan, np.nan), (np.nan, np.nan), (np.nan, np.nan)))
    
    # size parameters    
    Nscale, N1, N2 = imCoefs.shape
     
    # choose the windows (the weigths) for average  
    y, x = np.ogrid[-L:L+1, -L:L+1]     
    d = np.sqrt( x**2 + y**2 ) /L
    if np.char.equal(window,'flatsquare'):
        W = np.ones((2*L+1,2*L+1))
    elif np.char.equal(window,'bisquare'):  
        W = bisquarekernel(d)
    elif np.char.equal(window,'flatcircle'):
        W=np.copy(d)
        W[d<=1]=1
        W[d>1]=0    
    else:
        W = np.ones((2*L+1,2*L+1))
        print('Not yet implemented --> take flat window instead\n')
    Wgrid = np.zeros((2*L+1,2*L+1))
    Wgrid[d<gridsize/L]=1
    Nk = 2*L+1
    Nfft1 = N1+Nk-1
    Nfft2 = N2+Nk-1
    if len(fftplan)==0:
        fft_object,  ffti_object = prepare_plan2d(Nfft1, Nfft2, N_threads)
    else:
        fft_object, ffti_object = fftplan
   
    TFW=np.copy(fft_object(np.pad(Wgrid, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
    Coefs1 = np.copy(imCoefs[0,:,:])
    iNan=np.isnan(Coefs1)
    imtmp=np.ones((N1,N2))
    imtmp[iNan]=0
    imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
    CountValid=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
    Index0 = CountValid < 0.5
    
    # fft of the window
    TFW=np.copy(fft_object(np.pad(W, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
   
    
    # results allocations
    Mean_Std_loc = np.zeros((2, Nscale, N1, N2))*np.nan
    Mom12_loc = np.zeros((2, Nscale, N1, N2))*np.nan
    Cum12_loc = np.zeros((2, Nscale, N1, N2))*np.nan
    Cum1leader_loc = np.zeros((Nscale, N1, N2))*np.nan

    Mom4_loc = np.zeros((Nscale, N1, N2))*np.nan
    Flatness = np.zeros((Nscale, N1, N2))*np.nan

     
    # loop on scales
    for i in range(Nscale):
        print(i,end=' ',flush=True)
        Coefs1 = np.copy(imCoefs[i,:,:])
        iNan=np.isnan(Coefs1)
        imtmp=np.ones((N1,N2))
        imtmp[iNan]=0
        imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
        CountValid[CountValid < 0.1] = np.nan # must be integer >0
                    
        Coefs1[iNan]=0
        # compute moments and flatness
        imtmp=np.pad(Coefs1, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Mean_Std_loc[0,i,:,:]=lasum/CountValid
        lasum=convFFTW(np.abs(imtmp), TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Mom12_loc[0,i,:,:]=lasum/CountValid
        lasum=convFFTW(imtmp**2, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Mom12_loc[1,i,:,:]=lasum/CountValid
        Mean_Std_loc [1,i,:,:]= Mom12_loc[1,i,:,:] - Mean_Std_loc [0,i,:,:]**2
        lasum=convFFTW(imtmp**4, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Mom4_loc[i,:,:]=lasum/CountValid
        
        Flatness[i,:,:]= Mom4_loc[i,:,:]/(3*Mom12_loc[1,i,:,:]**2)
        
        # normalisation
        LCoefs1 = np.copy(imLCoefs[i,:,:])
        iNan=np.isnan(LCoefs1)
        imtmp=np.ones((N1,N2))
        imtmp[iNan]=0
        imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
        CountValid[CountValid < 0.1] = np.nan # must be integer >0
        #
        ####ITT compute C1 from log coeff
        logCoefs = np.copy(Coefs1)
        logCoefs = np.log(np.abs(logCoefs))
        logCoefs[np.isinf(logCoefs)] = np.nan
        iNanLC=np.isnan(logCoefs)
        logCoefs[iNanLC] =0
        imtmp=np.pad(logCoefs, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Cum12_loc[0,i,:,:]=lasum/CountValid       
        
        #
        logLCoefs = np.copy(LCoefs1)
        logLCoefs = np.log(np.abs(logLCoefs))
        logLCoefs[np.isinf(logLCoefs)] = np.nan
        logLCoefs[np.isnan(logLCoefs)] = 0
        imtmp=np.pad(logLCoefs, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Cum1leader_loc[i,:,:]=lasum/CountValid
        #Cum12_loc[0,i,:,:]=lasum/CountValid  
        lasum=convFFTW(imtmp**2, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0]=np.nan
        Cum12_loc[1,i,:,:]=lasum/CountValid-Cum1leader_loc[i,:,:]**2
        #Cum12_loc[1,i,:,:]=lasum/CountValid-Cum12_loc[0,i,:,:]**2
        
    print('.')
    
    Mean_Std_loc = Mean_Std_loc[:,:,L:-L,L:-L]
    Mom12_loc = Mom12_loc[:,:,L:-L,L:-L]
    Cum12_loc = Cum12_loc[:,:,L:-L,L:-L]
    Flatness = Flatness[:,L:-L,L:-L]
    
    # swap axes
    Mean_Std_loc = Mean_Std_loc.transpose(0,2,3,1)
    Mom12_loc = Mom12_loc.transpose(0,2,3,1)
    Cum12_loc = Cum12_loc.transpose(0,2,3,1)
    Flatness = Flatness.transpose(1,2,0)
    return Mean_Std_loc, Mom12_loc, Flatness ,Cum12_loc
#%%
def LocalCorrAnalysisRD(Coefs1sig,Coefs2sig,L, window='flatcircle',fftplan=[], gridsize=1):
    '''
        Compute localized correlation coeficients between two sets of values.
        
        rho = LocalCorrAnalysisRD(Coefs1,Coefs2,L, window)
    
    
        Inputs :
            
            coefs1  : first set of ceoficients 
            
            coefs2  : second set of ceoficients 
           
            L : size of the average (integer stricty positif)
           
            window : weighting function to use. Default : window='flat'.
                     can be window='bisquare'
           
       Outputs :   

            rho : localized correlation coeficients
           
    '''
    
    assert(Coefs1sig.shape == Coefs2sig.shape)
    # transpose axis
    Coefs1sig = Coefs1sig.transpose(2, 0, 1)
    Coefs2sig = Coefs2sig.transpose(2, 0, 1)
    
    # take care of border effect --> pad by nan nan 
    imCoefs1 = np.pad(Coefs1sig, ((0, 0), (L, L) ,(L, L)), constant_values=((np.nan, np.nan), (np.nan, np.nan), (np.nan, np.nan)))
    imCoefs2 = np.pad(Coefs2sig, ((0, 0), (L, L) ,(L, L)), constant_values=((np.nan, np.nan), (np.nan, np.nan), (np.nan, np.nan)))
    
    # size parameters       
    Nscale, N1, N2 = imCoefs1.shape
     
    # choose the windows (the weigths) for average  
    y, x = np.ogrid[-L:L+1, -L:L+1]     
    d = np.sqrt( x**2 + y**2 ) /L
    
    if np.char.equal(window,'flat'):
        W = np.ones((2*L+1,2*L+1))
        #y, x = np.ogrid[-L:L+1, -L:L+1]     
        #d = np.sqrt( x**2 + y**2 ) /L
    elif np.char.equal(window,'bisquare'):
        #y, x = np.ogrid[-L:L+1, -L:L+1]     
        #d = np.sqrt( x**2 + y**2 ) /L
        W = bisquarekernel(d)
    elif np.char.equal(window,'flatcircle'):
        W=np.copy(d)
        W[d<=1]=1
        W[d>1]=0    
        ###ITT
    else:
        #y, x = np.ogrid[-L:L+1, -L:L+1]     
        #d = np.sqrt( x**2 + y**2 ) /L
        W = np.ones((2*L+1,2*L+1))
        print('Not yet implemented --> take flat window instead\n')
     
    # for grid size
    Wgrid = np.zeros((2*L+1,2*L+1))
    Wgrid[d<gridsize/L]=1
    
    # fftw plan
    Nk = 2*L+1
    Nfft1 = N1+Nk-1
    Nfft2 = N2+Nk-1
    if len(fftplan)==0:
        fft_object,  ffti_object = prepare_plan2d(Nfft1, Nfft2, N_threads)
    else:
        fft_object, ffti_object = fftplan

    TFW=np.copy(fft_object(np.pad(Wgrid, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))
    Coefs1 = np.copy(imCoefs1[0,:,:])
    iNan=np.isnan(Coefs1)
    imtmp=np.ones((N1,N2))
    imtmp[iNan]=0
    imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
    CountValid=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
    Coefs2 = np.copy(imCoefs2[0,:,:])
    iNan=np.isnan(Coefs1)
    imtmp=np.ones((N1,N2))
    imtmp[iNan]=0
    imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
    CountValid2=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
    Index0=(CountValid < 0.5) & (CountValid2 < 0.5)
    
   
    TFW=np.copy(fft_object(np.pad(W, ((0, Nfft1-Nk) ,(0, Nfft2-Nk)), constant_values=((0, 0),(0,0)))))  
    rho = np.zeros((Nscale, N1, N2))*np.nan
    
    for i in range(Nscale):
        print(i,end=' ',flush=True)
        Coefs1 = np.copy(imCoefs1[i,:,:])
        Coefs2 = np.copy(imCoefs2[i,:,:])
        
        iNan1 = np.isnan(Coefs1)
        iNan2 = np.isnan(Coefs2)
        
        
        # Normalisation
        imtmp=np.ones((N1,N2))
        imtmp[iNan1]=0
        imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid1=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
        CountValid1[CountValid1 < 1e-5] = np.nan # must be integer >0
        imtmp=np.ones((N1,N2))
        imtmp[iNan2]=0
        imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid2=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
        CountValid2[CountValid2 < 1e-5] = np.nan # must be integer >0
        imtmp=np.ones((N1,N2))
        imtmp[iNan1]=0
        imtmp[iNan2]=0
        imtmp=np.pad(imtmp, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))
        CountValid12=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]       
        CountValid12[CountValid12 < 1e-5] = np.nan # must be integer >0
        
        Coefs1[iNan1]=0
        Coefs2[iNan2]=0
        imtmp=np.pad(Coefs1, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0] = np.nan
        moy1=lasum/CountValid1
        
        lasum=convFFTW(imtmp**2, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0] = np.nan
        M2_1=lasum/CountValid1
        var1 = M2_1 - moy1**2    
        var1[var1< 1e-5]=np.nan  

        imtmp=np.pad(Coefs2, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0] = np.nan
        moy2=lasum/CountValid2
        lasum=convFFTW(imtmp**2, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        M2_2=lasum/CountValid2
        var2 = M2_2 - moy2**2   
        var2[var2< 1e-5]=np.nan  
        imtmp=np.pad(Coefs1*Coefs2, ((0, Nk-1) ,(0, Nk-1)  ), constant_values=((0, 0),(0,0)))      
        lasum=convFFTW(imtmp, TFW, fft_object,ffti_object)[L:Nfft1-L,L:Nfft2-L]
        lasum[Index0] = np.nan
        moy_cross=lasum/CountValid12
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            rho[i,:,:] = (moy_cross - moy1*moy2)/np.sqrt(var1*var2)
           
    print('.')
    rho = rho[:,L:-L,L:-L]
    # swap axes
    rho = rho.transpose(1,2,0)
    return rho



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan  9 10:39:53 2023

@author: J. Lengyel, January 2023, janka.lengyel@ens-lyon.fr
"""

from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    readme = fh.read()

setup(
   name='lompy',
   version='0.1.12',
   author='S.G. Roux, J. Lengyel',
   author_email='jankalengyel@gmail.com',
   packages=find_packages(),
   url='https://gitlab.com/sroux67/LomPy',
   license='GNU AGPLv3',
   description='Local Multiscale Analysis of Marked Point Processes',
   long_description = readme,
   install_requires=['pyfftw','numpy','geopandas','matplotlib'],
   platforms=['any'],
   entry_points={"console_scripts": ["lompy = lompy.lompy:main"]},
   #long_description=open('README.txt').read(),
   #install_requires=[],
)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Feb  2 12:57:01 2023

@author: jankalengyel
"""

# __all__ = ['knn', 'conv2d','aux','viz']
# from knn import knn
# from conv2D import conv2D
# from aux import aux
# from viz import viz

from .lompy import *